# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from typing import (
    Any,
    Dict,
)

from qctrl.builders.graphql_utils.handlers import NamedObjectHandler
from qctrl.builders.type_registry import TypeOverrideMixin


class ComplexNumberHandler(
    NamedObjectHandler, TypeOverrideMixin
):  # pylint:disable=too-many-ancestors
    """Custom handler for ComplexNumber."""

    _type_hint = complex
    _object_name = "ComplexNumber"

    def process_loaded_data(self, data: Dict[str, Any]):
        return complex(data.get("real") or 0, data.get("imag") or 0)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
import time
from bisect import bisect
from collections import namedtuple
from typing import List

LOGGER = logging.getLogger(__name__)


class Wait:  # pylint:disable=too-few-public-methods
    """Class that can be used to sleep the program differing amounts of time
    based on ranges. Range must start at 0 and must be specified in ascending
    order.
    """

    Interval = namedtuple("Interval", ["frequency", "start"])
    default_intervals = [
        Interval(frequency=2, start=0),
        Interval(frequency=5, start=60 * 2),
        Interval(frequency=10, start=60 * 10),
    ]

    def __init__(self, intervals: List[Interval] = None):
        self._elapsed_time = 0
        self._intervals, self._frequency = self._build(
            intervals or self.default_intervals
        )

    @staticmethod
    def _build(intervals: List[Interval]):

        _frequency, _intervals = zip(*intervals)

        if any(val < 0 for val in _frequency):
            raise ValueError("Frequency values cannot be negative.")

        if any(val < 0 for val in _intervals):
            raise ValueError("Interval range values cannot be negative.")

        if any(_intervals[i] > _intervals[i + 1] for i in range(len(_intervals) - 1)):
            raise ValueError("Interval ranges need to be ascending.")

        # builds an updated frequency list to be used with bisection
        _frequency_list = [_frequency[0], *list(_frequency), _frequency[-1]]
        return _intervals, _frequency_list

    def _get_frequency(self, value: int):
        """
        Uses bisection to retrieve the frequency.
        """
        num = bisect(self._intervals, value)
        return self._frequency[num]

    def __call__(self):
        _current_frequency = self._get_frequency(self._elapsed_time)
        LOGGER.debug("Sleeping for %s", _current_frequency)
        time.sleep(_current_frequency)
        self._elapsed_time += _current_frequency



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from typing import (
    Dict,
    Set,
)

from dateutil.parser import isoparse
from gql import gql
from inflection import underscore
from qctrlcommons.exceptions import QctrlException
from rich.console import Console
from rich.table import Table

from .base import StaticQuery


class ActivityMonitorQuery(StaticQuery):
    """Retrieves recent activity."""

    query = gql(
        """
        query getActions($limit: Int, $offset: Int, $filterBy: ActionFilter) {
            actions(limit:$limit, offset:$offset, filterBy:$filterBy) {
                actions {
                    name
                    status
                    modelType
                    progress
                    createdAt
                    updatedAt
                    modelId
                }
                errors {
                    message
                }
            }
        }
    """
    )

    def _get_variable_values(  # pylint:disable=arguments-differ
        self, limit: int, offset: int, status: str = None, action_type: str = None
    ):

        if limit < 1:
            raise QctrlException("Limit cannot be less than 1.")

        if offset < 0:
            raise QctrlException("Offset cannot be less than 0.")

        if status:
            valid_statuses = self.get_valid_statuses()

            if status not in valid_statuses:
                raise QctrlException(
                    f"Status '{status}' is not valid. "
                    f"Please choose from a valid status type: "
                    f"{valid_statuses}"
                )

        if action_type:
            valid_action_types = self.get_valid_types()

            if action_type not in valid_action_types:
                raise QctrlException(
                    f"Action type '{action_type}' is not valid. "
                    f"Please choose from a valid action type: "
                    f"{valid_action_types}"
                )

        filter_by = {}

        if status:
            filter_by["status"] = {"exact": status}

        if action_type:
            filter_by["modelType"] = {"exact": action_type}

        return {"limit": limit, "offset": offset, "filterBy": filter_by}

    def _format_response(self, response: Dict, *_) -> None:
        def _snake_core_name(name):
            if name[:6] == "core__":
                name = underscore(name[6:])
            return name

        table = Table(title="Activity monitor", show_lines=True)
        console = Console()

        table.add_column("Action name\n(Type)", overflow="fold")
        table.add_column("ID", overflow="fold")
        table.add_column("Status", justify="center")
        table.add_column("Created at\nCompleted at")
        table.add_column("Runtime", overflow="fold")

        for row in response["actions"]["actions"]:
            name = _snake_core_name(row["name"])
            created_at = isoparse(row["createdAt"]).strftime("%Y-%m-%d %H:%M:%S")
            completed_at = (
                ""
                if row["status"] == "STARTED" or row["status"] == "PENDING"
                else isoparse(row["updatedAt"]).strftime("%Y-%m-%d %H:%M:%S")
            )
            runtime = isoparse(row["updatedAt"]) - isoparse(row["createdAt"])

            table.add_row(
                f'{name}\n({row["modelType"]})',
                f'{row["modelId"]}',
                f'{row["status"]}\n({row["progress"]:.0%}) ',
                f"{created_at}\n{completed_at} ",
                f"{str(runtime).split('.', maxsplit=1)[0]} ",
            )

        console.print(table)

    def _get_enum_values(self, enum_type: str) -> Set:
        """Retrieves the enum values from the schema.
        Parameters
        ----------
        enum_type: str
            Graphql enum type name
        Returns
        -------
        Set
            enum set value.
        Raises
        ------
        TypeError
            invalid enum type name.
        """
        field_type = self._client.schema.get_type(enum_type)

        if field_type is None:
            raise TypeError(f"unknown enum_type: {enum_type}")

        return set(field_type.values.keys())

    def get_valid_statuses(self) -> Set:
        """Returns the set of valid statuses."""
        return self._get_enum_values("ActionStatusEnum")

    def get_valid_types(self) -> Set:
        """Returns the set of valid types."""
        return self._get_enum_values("ActionTypeEnum")



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import csv
import hashlib
import json
import logging
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Union,
)

import click
from qctrlcommons.exceptions import QctrlException

DEFAULT_AUTH_DIR = Path.home() / ".config" / "qctrl"
LOGGER = logging.getLogger(__name__)


def file_path_for_url(url: str) -> Path:
    """Returns a Path() instance for the default file location using the URL
    MD5 as the filename.

    Parameters
    ----------
    url: str
        path to the file.


    Returns
    -------
    Path
        path object.
    """
    file_name = hashlib.md5(url.encode()).hexdigest()
    return DEFAULT_AUTH_DIR / file_name


TokenType = Dict[str, Union[str, List[str], int]]


def write_auth_file(token: TokenType, file_path: Path) -> None:
    """Writes the authentication file to the specified Path.

    Parameters
    ----------
    token: TokenType
        OIDC token.
    file_path: Path
        path to qctrl credential.

    Raises
    ------
    QctrlException
        no permission to credential file.
    """

    try:
        file_path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
        file_path.touch(mode=0o600, exist_ok=True)
        file_path.write_text(json.dumps(token))
    except IOError as exc:
        LOGGER.error("%s", exc, exc_info=True)
        raise QctrlException("incorrect permissions for credentials file") from exc


def _write_action_result_to_csv(path: str, actions: List[Dict[str, Any]]):
    """
    Covert action result to csv format

    Parameters
    ----------
    path: str
        Saved file path. It requires filename and file path (i.e path/filename.csv).
    actions: List
        List of action data.
    """
    with open(path, "w", encoding="utf-8") as file:
        writer = csv.DictWriter(file, fieldnames=actions[0].keys())
        writer.writeheader()
        for json_dict_result in actions:
            json_dict_result["user"] = json_dict_result["user"]["username"]
            writer.writerow(json_dict_result)


def _process_output(result: Dict, output: str = None) -> None:
    """
    Presents the gql response. The result would be printed or save to
    output file with indentation.

    Parameters
    ----------
    result: Dict
        gql response.
    output: str
        output file (Default value = None).
    """
    if output:
        with open(output, "w", encoding="utf-8") as file:
            json.dump(result, file, indent=4, sort_keys=True)
            click.echo(f"The result has written to the output file: {output}")
    else:
        # print with indentation
        click.echo(json.dumps(result, indent=4, sort_keys=True))



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from typing import Union

from gql import gql

from .base import StaticQuery

LOGGER = logging.getLogger(__name__)


class GetMutationNameQuery(StaticQuery):  # pylint:disable=too-few-public-methods
    """This class is used to retrieve the results of a previously run `function`.
    If the function is still running it will wait until it's finished
    before returning the results."""

    query = gql(
        """
        query getMutationName($modelId: String!) {
            action(modelId: $modelId) {
                action {
                    ... on CoreAction {
                        mutationName
                    }
                }
                errors {
                    message
                }
            }
        }
    """
    )

    def _get_variable_values(
        self, action_id: Union[str, int]
    ):  # pylint:disable=arguments-differ
        return {"modelId": str(action_id)}

    def _format_response(self, response: dict, *_) -> str:
        return response["action"]["action"]["mutationName"]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Tests for qctrlauth.session module.
"""

import json
import time

import pytest
from jwt.api_jwt import encode
from oauthlib.oauth2.rfc6749.errors import TokenExpiredError
from qctrlcommons.exceptions import QctrlException

from qctrl.qctrlauth.session import (
    QctrlApplicationClient,
    QctrlOAuth2Session,
)


class Called(Exception):
    """Custom exception to test if mocked functions are called"""


def test_session_default_parameters():
    session = QctrlOAuth2Session()
    assert session
    assert not session.token

    with pytest.raises(ValueError):
        _ = session.access_token


@pytest.mark.parametrize(
    "parameters, exception, match",
    [
        ({"redirect_uri": "a"}, ValueError, "redirect_uri_port is required"),
        ({"redirect_uri_port": 1}, ValueError, "redirect_uri is required"),
    ],
)
def test_session_parameters_validations(parameters, exception, match):
    with pytest.raises(exception, match=match):
        _ = QctrlOAuth2Session(**parameters)


def test_auto_generating_urls():
    session = QctrlOAuth2Session(base_url="http://test")
    assert session.auth_url
    assert session.token_url
    assert session.auto_refresh_url
    assert session.user_info_url


def test_session_authentication_needed(mocker):
    def _complete_login(*_):
        raise Called("_complete_login")

    mocker.patch("qctrl.qctrlauth.session.complete_login", new=_complete_login)

    session = QctrlOAuth2Session(base_url="http://test")
    with pytest.raises(Called, match="_complete_login"):
        session.authenticate_if_needed()


def test_session_authentication_not_needed(mocker):
    def _user_info(*_):
        return True

    mocker.patch(
        "qctrl.qctrlauth.session.QctrlOAuth2Session.user_info", new_callable=_user_info
    )

    valid_token = {"access_token": "foo", "expires_at": time.time() + 1000}
    session = QctrlOAuth2Session(base_url="http://test", token=valid_token)

    assert session.token is valid_token
    session.authenticate_if_needed()
    assert session.token is valid_token


def test_session_authentication_needed_with_valid_token(mocker):
    """
    Test cases where the access token looks valid but is not accepted by the server.
    """

    def _user_info(*_):
        return False

    def _complete_login(*_):
        raise Called("_complete_login")

    mocker.patch(
        "qctrl.qctrlauth.session.QctrlOAuth2Session.user_info", new_callable=_user_info
    )
    mocker.patch("qctrl.qctrlauth.session.complete_login", new=_complete_login)

    valid_token = {"access_token": "foo", "expires_at": time.time() + 1000}
    session = QctrlOAuth2Session(base_url="http://test", token=valid_token)

    with pytest.raises(Called, match="_complete_login"):
        session.authenticate_if_needed()


def test_session_non_expired_token():
    expires_at = time.time() + QctrlApplicationClient.REFRESH_THRESHOLD + 1
    session = QctrlOAuth2Session(
        token={"access_token": "foo", "expires_at": expires_at}
    )
    assert session.token
    assert session.access_token == "foo"


def test_session_with_expired_token_and_no_refresh():
    expires_at = time.time() + QctrlApplicationClient.REFRESH_THRESHOLD
    session = QctrlOAuth2Session(
        token={"access_token": "foo", "expires_at": expires_at}
    )
    assert session.token
    with pytest.raises(TokenExpiredError):
        _ = session.access_token


def test_session_refresh_token(mocker):
    def _refresh_token(*_):
        return {"access_token": "refreshed", "expires_at": time.time() + 1000}

    mocker.patch(
        "qctrl.qctrlauth.session.QctrlOAuth2Session.refresh_token", new=_refresh_token
    )
    expires_at = time.time() + QctrlApplicationClient.REFRESH_THRESHOLD
    original_token = {"access_token": "foo", "expires_at": expires_at}
    session = QctrlOAuth2Session(base_url="http://test", token=original_token)
    assert session.token is original_token
    assert session.access_token == "refreshed"


def test_session_token_persistance_new(tmp_path, mocker):
    file_path = tmp_path / "token.json"

    def _complete_login(_, __, callback):
        callback("response")

    def _fetch_token(self, *_, **__):
        self.token = {"access_token": "new", "expires_at": time.time() + 1000}
        return self.token

    mocker.patch("qctrl.qctrlauth.session.complete_login", new=_complete_login)
    mocker.patch("qctrl.qctrlauth.session.OAuth2Session.fetch_token", new=_fetch_token)

    session = QctrlOAuth2Session(base_url="http://test", session_file_path=file_path)
    with pytest.raises(ValueError, match="Missing access token"):
        _ = session.access_token
    session.authenticate_if_needed()
    assert session.access_token == "new"
    assert file_path.exists()
    assert "new" in file_path.read_text()


def test_session_token_persistance_existing(tmp_path, mocker):
    file_path = tmp_path / "token.json"

    def _refresh_token(*_):
        return {"access_token": "refreshed", "expires_at": time.time() + 1000}

    mocker.patch(
        "qctrl.qctrlauth.session.QctrlOAuth2Session.refresh_token", new=_refresh_token
    )
    expires_at = time.time() + QctrlApplicationClient.REFRESH_THRESHOLD
    original_token = {"access_token": "foo", "expires_at": expires_at}
    file_path.write_text(json.dumps(original_token))
    session = QctrlOAuth2Session(base_url="http://test", session_file_path=file_path)
    assert session.token == original_token
    assert session.access_token == "refreshed"


def test_session_token_persistance_invalid_path(tmp_path):
    file_path = tmp_path  # using a directory instead of a file
    with pytest.raises(QctrlException, match="cannot be a directory"):
        _ = QctrlOAuth2Session(base_url="http://test", session_file_path=file_path)


@pytest.mark.parametrize("user_role", ["cli-access", ""])
def test_has_role(mocker, user_role):
    def _user_info(*_):
        return True

    mocker.patch(
        "qctrl.qctrlauth.session.QctrlOAuth2Session.user_info", new_callable=_user_info
    )
    client_id = "boulder-opal-python"
    payload = {
        "exp": "898798798",
        "sub": "test_user_id",
        "realm_access": {"roles": ["delete"]},
        "resource_access": {client_id: {"roles": [user_role]}},
    }
    access_token = encode(payload, key="test")
    valid_token = {"access_token": access_token, "expires_at": time.time() + 1000}
    session = QctrlOAuth2Session(
        client_id=client_id, base_url="http://test", token=valid_token
    )
    assert session.has_role("cli-access") == bool(user_role)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging

LOGGER = logging.getLogger(__name__)


class ResultMixin:
    """Adds functionality to GraphQL result types."""

    # action statuses
    PENDING = "PENDING"
    STARTED = "STARTED"
    RETRY = "RETRY"
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"

    @property
    def action_id(self):
        """The id of the action."""
        return self.action.model_id

    @property
    def name(self):
        """The name of the action."""
        return self.action.name

    @property
    def status(self):
        """The status of the action."""
        return self.action.status

    @property
    def job_errors(self):
        """Any errors while executing the action."""
        return self.action.errors

    @property
    def progress(self):
        """The progress of the action."""
        return self.action.progress if self.action else 0

    @property
    def is_completed(self) -> bool:
        """Checks if action is completed."""
        if not self.status:
            return False

        return self.status not in (self.PENDING, self.STARTED, self.RETRY)

    @property
    def is_successful(self) -> bool:
        """Checks if action is successful."""
        return self.status == self.SUCCESS



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

"""Module that defines CLI scripts to be used from installed pip package.

For example:
    $ qctrl auth
"""
import logging
import os
from pathlib import Path
from typing import Any

import click
import inflection
from gql import gql
from qctrlcommons.exceptions import QctrlException

from qctrl import __version__
from qctrl.builders.client_builder import (
    create_client_auth,
    create_gql_client,
)
from qctrl.constants import (
    DEFAULT_API_ROOT,
    DEFAULT_OIDC_URL,
    OIDC_CLIENT_ID,
)
from qctrl.qctrlauth.session import QctrlOAuth2Session
from qctrl.queries import (
    ActivityMonitorQuery,
    GetQueueInfoQuery,
    GetWorkerInfoQuery,
)
from qctrl.queries.action_export import ActionExportQuery
from qctrl.scripts_utils import (
    TokenType,
    _process_output,
    _write_action_result_to_csv,
    file_path_for_url,
    write_auth_file,
)
from qctrl.utils import error_handler

DEFAULT_AUTH_DIR = Path.home() / ".config" / "qctrl"
LOGGER = logging.getLogger(__name__)


@click.group()
def main():
    """Q-CTRL CLI tool."""


@main.command()
@click.option("--access-token", required=True, help="JWT Access Token.")
@click.option("--refresh-token", required=True, help="JWT Refresh Token.")
@click.option(
    "--api-root",
    default=DEFAULT_API_ROOT,
    help="Custom Q-CTRL API base URL.",
    show_default=f"{DEFAULT_API_ROOT}",
)
@click.option(
    "--path",
    type=Path,
    default=lambda: os.environ.get("QCTRL_AUTHENTICATION_CREDENTIALS"),
    help=(
        "Use this option to set a custom location for the authentication files. "
        "If preset, will use the ENV variable `QCTRL_AUTHENTICATION_CREDENTIALS`. "
        "Otherwise, defaults to the standard config location."
    ),
)
def generate_auth_file(
    token: TokenType, oidc_url: str = DEFAULT_OIDC_URL, path: Path = None
) -> Any:
    """Generates the Q-CTRL Authentication file from command line.

    Parameters
    ----------
    token: str
        OIDC token.
    oidc_url: str, optional
        qctrl oidc url: (Default value = "https://id.q-ctrl.com/")
    path: Path
        (Default value = None)
    Returns
    -------
    """
    if not path:
        path = file_path_for_url(oidc_url)

    write_auth_file(token, path)


@main.command()
@click.option(
    "--oidc-url",
    default=DEFAULT_OIDC_URL,
    help="Q-CTRL OIDC base URL.",
    show_default=f"{DEFAULT_OIDC_URL}",
)
@click.option(
    "--path",
    type=Path,
    default=lambda: os.environ.get("QCTRL_AUTHENTICATION_CREDENTIALS"),
    help=(
        "Use this option to set a custom location for the authentication files. "
        "If preset, will use the ENV variable `QCTRL_AUTHENTICATION_CREDENTIALS`. "
        "Otherwise, defaults to the standard config location."
    ),
)
@error_handler
def auth(oidc_url: str = DEFAULT_OIDC_URL, path: str = None) -> QctrlOAuth2Session:
    """Q-CTRL Authentication setup.

    Provides default configuration for environmental authentication,
    allowing users to avoid managing in-script credentials. After this setup,
    the Qctrl package can be used in a Python environment as below:

        \b
        $ python
        >>> from qctrl import Qctrl
        >>> qctrl = Qctrl()

    For more details, access our documentation at https://docs.q-ctrl.com
    """

    oidc_client = QctrlOAuth2Session(
        OIDC_CLIENT_ID,
        base_url=oidc_url,
        session_file_path=Path(path) if path else file_path_for_url(oidc_url),
    )
    oidc_client.authenticate()
    return oidc_client


@main.command()
@click.option(
    "--options",
    help="Shows a list of available options for a particular argument.",
    type=click.Choice(["status", "type"], case_sensitive=False),
)
@click.option(
    "--limit",
    help="The number of previously ran actions to show.",
    type=int,
    default=5,
    show_default=True,
)
@click.option(
    "--offset",
    help="Offset the list of actions by a certain amount.",
    type=int,
    default=0,
    show_default=True,
)
@click.option("--status", help="The status of the action.")
@click.option("--type", help="The action type.")
@click.option(
    "--api-root",
    default=DEFAULT_API_ROOT,  # pylint: disable=too-many-arguments
    help="Custom Q-CTRL API base URL.",
    show_default=f"{DEFAULT_API_ROOT}",
)
def activity(
    options, limit, offset, status, type, api_root
):  # pylint: disable=redefined-builtin
    """
    Shows previously run actions and their statuses.
    """
    # Attempts to use previously store auth file, prompts interactive auth otherwise.
    gql_api = create_gql_client(api_root, create_client_auth(api_root))

    activity_monitor = ActivityMonitorQuery(gql_api)

    if options:
        click.secho(
            f"The list of available {inflection.pluralize(options)} are: ", bold=True
        )
        if options == "status":
            [  # pylint: disable=expression-not-assigned
                click.secho(value) for value in activity_monitor.get_valid_statuses()
            ]
        if options == "type":
            [  # pylint: disable=expression-not-assigned
                click.secho(value) for value in activity_monitor.get_valid_types()
            ]
    else:
        activity_monitor(limit=limit, offset=offset, status=status, action_type=type)


@main.command(help="Displays the current version of the 'qctrl' package.")
def version() -> None:
    """
    Outputs the current package version.
    """
    click.secho(__version__)


@main.command(help="Describes the current environment.")
@click.option(
    "--api-root",
    default=DEFAULT_API_ROOT,  # pylint: disable=too-many-arguments
    help="Custom Q-CTRL API base URL.",
    show_default=f"{DEFAULT_API_ROOT}",
)
@click.option(
    "--type",
    help="specify the report type.",
    default="queue",
    type=click.Choice(["queue", "worker"], case_sensitive=False),
    show_default="queue",
)
def env(api_root, type):  # pylint:disable=redefined-builtin
    """Shows details to allow monitoring of the current
    environment.
    """
    gql_api = create_gql_client(api_root, create_client_auth(api_root))
    if type and type == "worker":
        # display per-worker view
        get_env = GetWorkerInfoQuery(gql_api)
    else:
        # display per-queue view
        get_env = GetQueueInfoQuery(gql_api)
    print("Fetching environment details. Please wait ...")
    click.secho(get_env())


@main.command(help="Run GraphQL query.")
@click.option(
    "--api-root",
    default=DEFAULT_API_ROOT,
    help="Custom Q-CTRL API base URL.",
    show_default=f"{DEFAULT_API_ROOT}",
)
@click.option("--input", "--i", help="input graphql file", required=True)
@click.option("--output", "--o", help="output file", default=None, show_default=None)
def gql_query(api_root, input, output):  # pylint:disable=redefined-builtin
    """
    Runs gql query and return indented result.
    """
    gql_api = create_gql_client(api_root, create_client_auth(api_root))
    with open(input, "r", encoding="utf-8") as file:
        query = file.read()
    result = gql_api.execute(gql(query))

    if result.get("errors"):
        # raise exception when there's any root level errors
        raise QctrlException(result["errors"])

    _process_output(result, output)


@main.command(help="Get JWT access token.")
@click.option(
    "--oidc-url",
    default=DEFAULT_OIDC_URL,
    help="Custom Q-CTRL OIDC base URL.",
    show_default=f"{DEFAULT_OIDC_URL}",
)
@click.option("--output", "--o", help="output file", default=None, show_default=None)
@click.pass_context
def access_token(ctx, oidc_url, output):
    """
    Get JWT access token.
    """
    oidc_session = ctx.invoke(auth, oidc_url=oidc_url)
    _process_output({"access_token": oidc_session.access_token}, output)


@main.command(help="Export the action data into csv file.")
@click.option(
    "--api-root",
    help="Custom Q-CTRL API base URL.",
    default=DEFAULT_API_ROOT,
    show_default=f"{DEFAULT_API_ROOT}",
)
@click.option(
    "--oidc-url",
    default=DEFAULT_OIDC_URL,
    help="Custom Q-CTRL OIDC base URL.",
    show_default=f"{DEFAULT_OIDC_URL}",
)
@click.option("--start-date", help="The date begin i.e 2020-12-31", required=True)
@click.option("--end-date", help="The date till i.e 2021-06-30", required=True)
@click.option(
    "--ignore-test-users",
    help="Flag to ignore the action records created by test users.",
    default=True,
    show_default=True,
)
@click.option(
    "--file-name",
    help="Saved file name",
    default="result.csv",
    show_default="result.csv",
)
@click.option("--path", help="Path of saved file", required=True)
def action_export(
    api_root, oidc_url, start_date, end_date, ignore_test_users, file_name, path
):
    """
    Exports action data to csv file.
    """
    directory_path = Path.home() / path
    if not directory_path.exists():
        click.echo(f"Cannot find Path {directory_path}")
    else:
        gql_api = create_gql_client(api_root, create_client_auth(oidc_url))
        action_export_query = ActionExportQuery(gql_api)
        result = action_export_query(start_date, end_date, ignore_test_users)

        if not result:
            click.echo("No match action data found")
        else:
            # format and save the result to file
            saved_path = directory_path / file_name
            _write_action_result_to_csv(saved_path, result)

            click.echo(f"The result has been saved to {saved_path}")



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from graphql import GraphQLNonNull

from .base import BaseHandler
from .mixins import (
    DataMixin,
    PayloadMixin,
    SelectionMixin,
)


class NonNullHandler(BaseHandler, PayloadMixin, SelectionMixin, DataMixin):
    """Handler for GraphQLNonNull types."""

    _field_type = GraphQLNonNull

    def _format_payload(self, field_type, data):
        result = self._env.format_payload(field_type.of_type, data)

        if result is None:
            # IMPROVEMENT: catch this error and include field
            # name in error message
            raise ValueError("non-null value expected")

        return result

    def _get_selection(self, field_type, overrides=None):
        return self._env.get_selection(field_type.of_type, overrides)

    def _load_data(self, field_type, data):
        return self._env.load_data(field_type.of_type, data)

    def _update_data(self, field_type, obj, data):
        return self._env.update_data(field_type.of_type, obj, data)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring

import logging
from typing import Dict
from urllib.parse import urljoin

from gql import Client
from qctrlcommons.auth import ClientAuthBase
from qctrlcommons.exceptions import QctrlException
from qctrlcommons.utils import generate_user_agent

from qctrl.builders.http_transport import (
    QctrlGqlClient,
    QctrlRequestsHTTPTransport,
)
from qctrl.constants import OIDC_CLIENT_ID
from qctrl.qctrlauth.session import QctrlOAuth2Session
from qctrl.utils import (
    check_client_version,
    error_handler,
)

from .. import __version__ as qctrl_version
from ..scripts_utils import file_path_for_url

LOGGER = logging.getLogger(__name__)


class OIDCClientAuth(ClientAuthBase):
    """
    Client authentication for OIDC.
    """

    def __init__(self, oidc_session):
        super().__init__()
        self.oidc_session = oidc_session

    def encode(self) -> str:
        return f"Bearer {self.oidc_session.access_token}"


@error_handler
def _create_oidc_session(oidc_url: str) -> QctrlOAuth2Session:
    """
    Returns the GQL client.

    Parameters
    ----------
    oidc_url: str
        The Api url the package should use.

    Returns
    -------
    QctrlOAuth2Session
        OpenID compliant OAuth2Session.
    """

    oidc_session = QctrlOAuth2Session(
        client_id=OIDC_CLIENT_ID,
        base_url=oidc_url,
        session_file_path=file_path_for_url(oidc_url),
    )
    oidc_session.authenticate_if_needed()
    return oidc_session


def create_client_auth(oidc_url: str) -> ClientAuthBase:
    """
    Returns the GQL client.

    Parameters
    ----------
    oidc_url: str
        The Api url the package should use.

    Returns
    -------
    ClientAuthBase
        Object to handle authentication in the GQL Client.

    Raises
    ------
    QctrlException
        If user doesn't have the `cli-access` role.
    """
    oidc_session = _create_oidc_session(oidc_url)
    if not oidc_session.has_role("cli-access"):
        oidc_session.invalidate_access_token()
        raise QctrlException(
            "Access not allowed due to invalid subscription. "
            "Please access `https://boulder.q-ctrl.com` to activate your access."
        )

    return OIDCClientAuth(oidc_session)


@check_client_version
def create_gql_client(
    api_root: str, auth: ClientAuthBase = None, headers: Dict = None
) -> Client:
    """Creates a gql client.

    Parameters
    ----------
    auth: ClientAuthBase
        The auth object.
    api_root: str
        The host url.
    headers: Dict
        HTTP headers


    Returns
    -------
    Client
        The GQL Client.
    """

    url = urljoin(api_root, "graphql/")
    LOGGER.debug("gql url: %s", url)
    headers = headers or {}

    headers.update(
        {
            "Content-Encoding": "gzip",
            "Content-Type": "application/json",
            "User-Agent": generate_user_agent("Q-CTRL Python", version=qctrl_version),
        }
    )

    # use_json needs to be set to false otherwise the library will attempt
    # to dump and encode the data twice.
    transport = QctrlRequestsHTTPTransport(
        url=url, use_json=False, headers=headers, auth=auth, retries=3
    )
    return QctrlGqlClient(transport=transport, fetch_schema_from_transport=True)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from typing import (
    Any,
    Dict,
)

from graphql import (
    GraphQLType,
    GraphQLUnionType,
)

from .base import BaseHandler
from .mixins import (
    DataMixin,
    SelectionMixin,
)


def _get_union_type_by_name(
    union_type: GraphQLUnionType, type_name: str
) -> GraphQLType:
    """
    Returns the type from a union by the name.

    Parameters
    ----------
    union_type: GraphQLUnionType
        A graphql union.
    type_name: str
        The name of the expected type to return.

    Raises
    ------
    TypeError
        When no type is found with the provided name.

    Returns
    -------
    GraphQLType
        The sub type from the union that matches the type name.
    """

    for sub_type in union_type.types:
        if sub_type.name == type_name:
            return sub_type

    raise TypeError(f"unhandled type {type_name} by {type(union_type)}")


class UnionHandler(BaseHandler, SelectionMixin, DataMixin):
    """
    Handler for GraphQLUnionType types.
    """

    _field_type = GraphQLUnionType

    def _get_selection(self, field_type, overrides=None):
        fields = ["__typename"]

        for field in field_type.types:
            nested_fields = self._env.get_selection(field)
            fields.append(f"... on {field.name} {nested_fields}")

        return self._env.format_query_selection_fields(fields)

    def _load_data(self, field_type, data):
        type_name = None

        if "__typename" not in data:
            type_name = self._find_type(field_type.name, data)
        else:
            type_name = data.pop("__typename")

        if type_name is None:
            raise TypeError(f"unresolved union type for field {field_type.name}")

        union_type = _get_union_type_by_name(field_type, type_name)
        return self._env.load_data(union_type, data)

    @staticmethod
    def _find_type(field_type: str, data: Dict[str, Any]):
        if field_type == "ComplexArray":
            if "dense_entries" in data:
                return "ComplexDenseArray"

            if "sparse_entries" in data:
                return "ComplexSparseArray"

        return None

    def _update_data(self, field_type, obj, data):
        union_type = _get_union_type_by_name(field_type, obj.__class__.__name__)
        return self._env.update_data(union_type, obj, data)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from typing import (
    Dict,
    List,
    Optional,
)

from gql import Client
from graphql import (
    DocumentNode,
    FieldNode,
    NameNode,
    ObjectValueNode,
    OperationDefinitionNode,
    SelectionSetNode,
    ValueNode,
    VariableDefinitionNode,
    VariableNode,
    print_ast,
)
from graphql.pyutils import FrozenList

from .base import BaseQuery

LOGGER = logging.getLogger(__name__)


class MultiQuery(BaseQuery):
    """Used to combine multiple queries into a
    single document.
    """

    name: str = None

    def __init__(self, client: Client, *queries: List[BaseQuery]):
        super().__init__(client)
        self._queries = list(queries)

    def add_query(self, query: BaseQuery):
        """Adds a query to be added to the
        executed document.
        """
        self._queries.append(query)

    def _get_query_document(self):

        if not self._queries:
            raise ValueError("no queries found")

        if len(self._queries) == 1:
            document = self._queries[0].get_query_document()
            document = self._transform_sub_query_document(0, document)
            return document

        documents = []

        for index, query in enumerate(self._queries):
            document = query.get_query_document()
            document = self._transform_sub_query_document(index, document)
            documents.append(document)

        combined_document = _merge_operation_documents(documents, self.name)
        LOGGER.debug("combined_document: %s", print_ast(combined_document))
        return combined_document

    def _transform_sub_query_document(  # pylint:disable=no-self-use
        self, index: int, document: DocumentNode  # pylint:disable=unused-argument
    ) -> DocumentNode:
        """Optional hook which can be used by
        a sub class to transform the sub query
        documents before merging.
        """
        return document

    def _format_response(self, response, document) -> List[Dict]:
        result = []

        for key in self._get_document_response_keys(document):

            if key not in response:
                raise ValueError(f"expected response key not found: {key}")

            result.append(response[key])

        return result


def _find_variable_nodes(value_node: ValueNode) -> List[VariableNode]:
    """Recursively collects VariableNode objects
    within the given value node.
    """
    result = []

    if isinstance(value_node, VariableNode):
        result.append(value_node)

    # can contain nested variable nodes
    if isinstance(value_node, ObjectValueNode):
        for object_field_node in value_node.fields:
            result += _find_variable_nodes(object_field_node.value)

    return result


def _suffix_document(doc: DocumentNode, suffix: str):
    """Attach a suffix to the variable names and retrieval
    keys for the given document. This function can be used
    in preparation for combining multiple instances of the
    same document.
    """
    variables: Dict[str, List[VariableNode]] = {}
    field_nodes: List[FieldNode] = []

    for definition_node in doc.definitions:

        for variable_definition_node in definition_node.variable_definitions:
            variable_name = variable_definition_node.variable.name.value

            if variable_name in variables:
                raise ValueError(
                    f"duplicate variable name in document: {variable_name}"
                )

            variables[variable_name] = [variable_definition_node.variable]

        for field_node in definition_node.selection_set.selections:

            if not isinstance(field_node, FieldNode):
                raise TypeError(f"unhandled node type: {field_node}")

            field_nodes.append(field_node)

            # find references to variables in field arguments
            for argument_node in field_node.arguments:

                for variable_node in _find_variable_nodes(argument_node.value):
                    variable_name = variable_node.name.value

                    if variable_name not in variables:
                        raise ValueError(f"undefined variable name: {variable_name}")

                    variables[variable_name].append(variable_node)

    # update alias for field nodes with suffix
    for field_node in field_nodes:
        if not field_node.alias:
            field_node.alias = NameNode(value=field_node.name.value)

        field_node.alias.value = f"{field_node.alias.value}_{suffix}"

    # update variables names with suffix, including references
    for variable_name, variable_nodes in variables.items():
        new_variable_name = f"{variable_name}_{suffix}"

        for variable_node in variable_nodes:
            variable_node.name.value = new_variable_name


def _check_unique_variables(variable_definition_nodes: List[VariableDefinitionNode]):
    """Checks that the variable name is unique across all
    variable definition nodes.
    """
    variable_names = set()

    for node in variable_definition_nodes:
        if node.variable.name.value in variable_names:
            raise ValueError(f"duplicate variable: {node.variable.name.value}")

        variable_names.add(node.variable.name.value)


def _check_unique_retrieval_keys(field_nodes: List[FieldNode]):
    """Checks that the retrieval key (name or alias) is unique
    across all field nodes.
    """
    keys = set()

    for node in field_nodes:
        key = node.alias.value if node.alias else node.name.value

        if key in keys:
            raise ValueError(f"duplicate retrieval key: {key}")

        keys.add(key)


def _merge_operation_documents(
    documents: List[DocumentNode], name: Optional[str] = None
) -> DocumentNode:
    """Combines multiple documents into a new, single document with
    multiple queries. Variable names must be unique across all
    documents. The retrieval key (name or alias) of all selections
    must be unique across all documents.

    Parameters
    ----------
    documents: List[DocumentNode]
        list of GraphQL documents to be merged.

    name: Optional[str]
        name of the combined document.

    Returns
    -------
    DocumentNode
        the combined GraphQL document.

    Raises
    ------
    ValueError
        any validation errors.
    TypeError
        unhandled GraphQL node types.
    """

    if not documents:
        raise ValueError("no documents provided")

    if len(documents) == 1:
        raise ValueError("unable to merge single document")

    operation_type = None  # used to check consistent operation type
    variable_definition_nodes = []
    field_nodes = []

    # for each doc to be merged
    for doc in documents:
        assert isinstance(doc, DocumentNode)

        for definition_node in doc.definitions:

            if not isinstance(definition_node, OperationDefinitionNode):
                raise TypeError(f"unhandled DefinitionNode: {definition_node}")

            if operation_type is None:
                operation_type = definition_node.operation

            elif definition_node.operation != operation_type:
                raise ValueError(
                    "unable to merge different operation types; found"
                    f" {definition_node.operation}, expected {operation_type}"
                )

            # collect variable definitions
            for variable_definition_node in definition_node.variable_definitions:
                variable_definition_nodes.append(variable_definition_node)

            # collect field nodes
            for selection_node in definition_node.selection_set.selections:
                if not isinstance(selection_node, FieldNode):
                    raise TypeError(f"unhandled SelectionNode: {selection_node}")

                field_nodes.append(selection_node)

    # validation for merged document
    _check_unique_variables(variable_definition_nodes)
    _check_unique_retrieval_keys(field_nodes)

    # build node for combined document
    operation_definition_node = OperationDefinitionNode(
        operation=operation_type,
        variable_definitions=FrozenList(variable_definition_nodes),
        selection_set=SelectionSetNode(selections=FrozenList(field_nodes)),
    )

    if name:
        operation_definition_node.name = NameNode(value=name)

    return DocumentNode(definitions=FrozenList([operation_definition_node]))



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from gql import gql

from .base import StaticQuery


class MeQuery(StaticQuery):  # pylint:disable=too-few-public-methods
    """Retrieves info for the current user."""

    query = gql(
        """
        query {
            me {
                me {
                    ... on User {
                        username
                    }
                }
                errors {
                    message
                    fields
                }
            }
        }
    """
    )

    def _format_response(self, response: dict, *_) -> str:
        return response["me"]["me"]["username"]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Functionality related to the Q-CTRL graph structure.

The commons objects are re-imported here to allow all access to the objects to happen directly from
the `qctrl` package.
"""

import sys

# We import Graph to expose it directly from this module.
# pylint: disable=unused-import
from qctrlcommons.graph import Graph as _BaseGraph
from qctrlcommons.node.registry import TYPE_REGISTRY as _TYPE_REGISTRY

from qctrl import toolkit as _toolkit
from qctrl.builders.namespaces import ToolkitCategory as _ToolkitCategory
from qctrl.builders.namespaces import build_and_bind_toolkit as _build_and_bind_toolkit
from qctrl.utils import PackageRegistry as _PackageRegistry

_module = sys.modules[__name__]
for _type_cls in _TYPE_REGISTRY:
    setattr(_module, _type_cls.__name__, _type_cls)

# pylint: disable=too-few-public-methods
class Graph(_BaseGraph):

    """
    Utility class for representing and building a Q-CTRL data flow graph.
    """

    def __init__(self):
        super().__init__()
        _build_and_bind_toolkit(self, _ToolkitCategory.NODES, _toolkit)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from .activity_monitor import ActivityMonitorQuery
from .get_environment import (
    GetQueueInfoQuery,
    GetWorkerInfoQuery,
)
from .get_mutation_name import GetMutationNameQuery
from .me import MeQuery
from .refresh_action import (
    MultiRefreshActionQuery,
    RefreshActionQuery,
)
from .start_action import (
    MultiStartActionQuery,
    StartActionQuery,
)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import inspect
import logging
from enum import Enum
from types import SimpleNamespace
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Union,
)

import inflection

from qctrl.dynamic import (
    BoulderModule,
    dynamic_class,
)

LOGGER = logging.getLogger(__name__)


FUNCTION_NAMESPACE_DOCSTRING = """
    Namespace for functions. Functions are computations using objects
    created from the `types` namespace.

    Boulder Opal also provides toolkits to simplify the workflow for common computation tasks.
    Some toolkits host functions that you can access from their namespaces of
    the :py:obj:`~qctrl.Qctrl` object.
    See the `reference <https://docs.q-ctrl.com/boulder-opal/references/qctrl/Toolkits.html>`_
    for details.
"""


TYPE_NAMESPACE_DOCSTRING = """
    Namespace for types. Objects created from these types are used
    when performing computations defined in the `functions` namespace.
"""


class BaseNamespace(SimpleNamespace):
    """Base namespace class for Qctrl components."""

    @classmethod
    def extend(cls, attrs: Dict[str, Any]) -> None:
        """Extends the namespace class by adding attributes.

        Parameters
        ----------
        attrs : Dict[str, Any]
            Dict of attributes to add to the namespace.

        Raises
        ------
        AttributeError
            Existing attribute found.
        """
        for attr, value in attrs.items():
            if hasattr(cls, attr):
                raise AttributeError(f"existing attr ({attr}) on namespace: {cls}")

            LOGGER.debug("adding attr %s to namespace: %s", attr, cls)
            setattr(cls, attr, value)

    @classmethod
    def extend_functions(cls, *funcs: Callable) -> None:
        """Extends the namespace class by adding functions as attributes. The
        function will be added as a staticmethod.

        Parameters
        ----------
        *funcs : Callable
            Functions to be added to the namespace.
        """
        for func in funcs:
            cls.extend({func.__name__: staticmethod(func)})


class TypeNamespaceMixin:
    """
    Mixin to handle namespaced types.
    """

    @classmethod
    def add_registry(cls, type_registry: "TypeRegistry"):
        """
        Adds the registered types as attributes of the namespace.
        """
        for name, type_cls in type_registry.get_type_map().items():
            attr, namespace = type_registry.parse_name(name)
            cls._namespace_extend(namespace or [], attr, type_cls)

        cls.registry = type_registry

    @classmethod
    def _namespace_extend(cls, namespace: List[str], attr: str, obj: Any):
        """
        Recursively create nested namespaces, then use extend
        to set the attribute.
        """
        current = cls
        prefix = ""  # name prefix for nested namespaces

        for item in namespace:
            nested_attr = inflection.underscore(item)
            prefix += inflection.camelize(nested_attr)

            # nested namespace already exists
            if hasattr(current, nested_attr):
                current = getattr(current, nested_attr)

            # created nested namespace
            else:
                nested_namespace_cls = _create_namespace_cls(prefix + "Type")
                nested_namespace = nested_namespace_cls()
                current.extend({nested_attr: nested_namespace})

                # update current for any further nesting
                current = nested_namespace_cls

        current.extend({attr: obj})


@dynamic_class("namespaces")
def _create_namespace_cls(
    prefix: str,
    docstring: Optional[str] = None,
    base: SimpleNamespace = BaseNamespace,
    mixins: Optional[List[type]] = None,
    module: BoulderModule = BoulderModule.QCTRL,
) -> SimpleNamespace:
    """Creates a new namespace class and returns an instance of it.

    Parameters
    ----------
    prefix : str
        A prefix for the all core namespace. (i.e core__ )
    docstring : str, optional
        Namespace docstring. Defaults to `None`.
    base : SimpleNamespace, optional
        Namespace base. Defaults to `BaseNamespace`.
    mixins : List[type], optional
        Nested mixins. Defaults to `None`.
    module : BoulderModule, optional
        The module to which the namespace belongs. Defaults to `BoulderModule.QCTRL`.

    Returns
    -------
    SimpleNamespace
        Formatted namespace.
    """
    name = f"{inflection.camelize(prefix)}Namespace"
    parents = [base]

    if mixins:
        parents += mixins

    attrs = {"__module__": module.value}

    if docstring:
        attrs.update({"__doc__": docstring})

    new_cls = type(name, tuple(parents), attrs)
    return new_cls


def create_function_namespace():
    """Creates the function namespace."""
    cls = _create_namespace_cls("Function", FUNCTION_NAMESPACE_DOCSTRING)
    return cls()


def create_type_namespace():
    """Creates the type namespace."""
    cls = _create_namespace_cls(
        "Type", TYPE_NAMESPACE_DOCSTRING, mixins=[TypeNamespaceMixin]
    )
    return cls()


class ToolkitCategory(Enum):
    """
    Defines the categories for toolkits.
    """

    FUNCTIONS = 1
    NODES = 2


def build_and_bind_toolkit(
    obj: Union["Qctrl", "Graph"], category: ToolkitCategory, toolkit: "qctrltoolkit"
) -> None:
    """
    Builds namespace for toolkit, add corresponding methods to
    the namespace, and bind the namespace to the object.

    Note that a given method could belong to multiple namespaces.

    Parameters
    ----------
    obj : Union["Qctrl", "Graph"]
        The object (either a qctrl or graph object) to which the toolkit is bound.
    category : ToolkitCategory
        The type of toolkit.
    toolkit : Any
        The toolkit module.
    """

    # ensure that attributes we need are correctly exposed
    _check_attrs(toolkit, ["TOOLKIT_ATTR", category.name, "forge_toolkit"])

    # create a set of labels for namespaces from toolkit
    namespace_labels = set()
    for method in getattr(toolkit, category.name):
        namespace_labels.update(
            (attr for attr in getattr(method, toolkit.TOOLKIT_ATTR))
        )

    # create namespace and bind it to the object
    for label in namespace_labels:
        _name = label.get_name()
        _namespace_cls = _create_namespace_cls(
            prefix=_name,
            docstring=label.get_doc(),
            module=BoulderModule.get_module(obj.__class__.__name__.lower()),
        )
        setattr(_namespace_cls, toolkit.TOOLKIT_ATTR, _name)

        _namespace = _namespace_cls()
        assert not hasattr(obj, _name), f"Namespace {_name} is already defined!"
        LOGGER.debug("adding attr to namespace %s to object %s", _namespace, obj)
        setattr(obj, _name, _namespace)

    # binds methods to namespaces
    for method in getattr(toolkit, category.name):
        forged = toolkit.forge_toolkit(method, obj)
        for namespace in getattr(method, toolkit.TOOLKIT_ATTR):
            _ns = getattr(obj, namespace.get_name())
            if inspect.isfunction(forged):
                _ns.extend_functions(forged)
            else:
                assert not hasattr(
                    _ns, method.__name__
                ), f"existing attr ({method}) on namespace: {_ns}"
                LOGGER.debug("adding attr %s to namespace: %s", method, _ns)
                setattr(_ns, method.__name__, forged)


def _check_attrs(package, attr_names: List[str]):
    for name in attr_names:
        assert hasattr(package, name), f"No module/attribute {name} found in {package}."


__all__ = [
    "BaseNamespace",
    "BoulderModule",
    "ToolkitCategory",
    "create_function_namespace",
    "create_type_namespace",
    "build_and_bind_toolkit",
]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from functools import lru_cache
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Union,
)

from graphql import GraphQLType

from .base_type_registry import BaseTypeRegistry
from .handlers import (
    BooleanScalarHandler,
    EnumHandler,
    FloatScalarHandler,
    IDScalarHandler,
    InputObjectHandler,
    IntScalarHandler,
    ListHandler,
    NonNullHandler,
    ObjectHandler,
    StringScalarHandler,
    UnionHandler,
)
from .handlers.base import BaseHandler
from .handlers.mixins import (
    DataMixin,
    PayloadMixin,
    SelectionMixin,
)

LOGGER = logging.getLogger(__name__)


class GraphQLEnvironment:
    """Defines an environment for using GraphQL
    type handlers.
    """

    _fallback_handlers = [
        NonNullHandler,
        FloatScalarHandler,
        IntScalarHandler,
        IDScalarHandler,
        BooleanScalarHandler,
        StringScalarHandler,
        ListHandler,
        InputObjectHandler,
        UnionHandler,
        ObjectHandler,
        EnumHandler,
    ]

    def __init__(
        self,
        type_registry_cls: BaseTypeRegistry = None,
        custom_handlers: List[BaseHandler] = None,
    ):
        self._type_registry = None

        if type_registry_cls:
            self._type_registry = type_registry_cls(self)

        custom_handlers = custom_handlers or []
        self._handlers = []

        for cls in custom_handlers + self._fallback_handlers:

            if not issubclass(cls, BaseHandler):
                raise TypeError(f"invalid handler class: {cls}")

            self._handlers.append(cls(self))

    @property
    def type_registry(self):  # pylint:disable=missing-function-docstring
        if self._type_registry is None:
            raise RuntimeError("No type registry defined")

        return self._type_registry

    def field_to_attr(self, field_name: str) -> str:  # pylint:disable=no-self-use
        """Converts a GraphQL field name into the
        corresponding attribute name if using objects
        when loading or updating data. By default, the
        attribute name is the same as the field name.
        """
        return field_name

    @lru_cache(maxsize=128)
    def get_handler(self, base: type, field_type: GraphQLType) -> BaseHandler:
        """
        Finds the correct handler for the field_type based on the base class
        provided.
        """
        for handler in self._handlers:
            if base.can_handle(handler, field_type):
                LOGGER.debug("using handler: %s", handler)
                return handler

        raise TypeError(f"unhandled type: {type(field_type)} ({field_type})")

    def iter_handlers(self, base: type) -> Iterator:
        """Iterates through the configured handlers
        that are instances of the base class provided.
        """
        for handler in self._handlers:
            if isinstance(handler, base):
                yield handler

    def format_payload(self, field_type: GraphQLType, data: Any) -> str:
        """Formats the given payload data for use in a query
        or mutation request.
        """
        LOGGER.debug("format_payload: %s, %s", field_type, data)

        handler = self.get_handler(PayloadMixin, field_type)
        return handler.format_payload(field_type, data)

    def get_selection(
        self, field_type: GraphQLType, overrides: Optional[Dict[str, str]] = None
    ) -> str:
        """Returns the selection query string for the given field type.
        By default, all nested fields will be included. Overrides can be
        provided to restrict which nested fields are selected.
        """
        LOGGER.debug("get_selection: %s", field_type)

        handler = self.get_handler(SelectionMixin, field_type)
        return handler.get_selection(field_type, overrides)

    def load_data(self, field_type: GraphQLType, data: Any) -> Any:
        """Loads the data from a GraphQL response into objects using
        the type registry.
        """
        LOGGER.debug("load_data: %s", field_type)

        handler = self.get_handler(DataMixin, field_type)
        return handler.load_data(field_type, data)

    def update_data(self, field_type: GraphQLType, obj: Any, data: Any) -> Any:
        """Updates an existing object (from the type registry) with
        response data.
        """
        LOGGER.debug("update_data: %s", field_type)

        handler = self.get_handler(DataMixin, field_type)
        return handler.update_data(field_type, obj, data)

    @staticmethod
    def get_nested_selection_override(
        overrides: Optional[Dict[str, str]], field_name: str
    ) -> Optional[Union[str, Dict[str, str]]]:
        """
        Returns the nested selection override for the
        given field name, if any.
        """
        result = None

        if isinstance(overrides, dict):
            try:
                result = overrides[field_name]
            except KeyError:
                pass

        return result

    def get_nested_value(self, obj: Any, field_name: str) -> Any:
        """Returns the nested value from the object
        for the given field name.
        """
        attr_name = self.field_to_attr(field_name)

        if isinstance(obj, dict):
            if field_name in obj:
                return obj[field_name]

            if attr_name in obj:
                return obj[attr_name]

        return getattr(obj, attr_name, None)

    @staticmethod
    def format_query_payload_dict(data: Dict[str, Any]) -> str:
        """Formats a key-value dictionary for a query."""

        values = [f"{key} : {value}" for key, value in data.items()]
        string_values = " ".join(values)

        return f"{{ {string_values} }}"

    @staticmethod
    def format_query_selection_fields(fields: List[str]) -> str:
        """Formats a list of values for a query."""
        # e.g '{ modelId name status errors progress }'
        string_fields = " ".join(fields)
        return f"{{ {string_fields} }}"



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring,too-many-ancestors
import json

from .base import BaseHandler
from .mixins import ScalarMixin


class FloatScalarHandler(ScalarMixin, BaseHandler):
    """Handler for GraphQL Float scalar type."""

    _scalar_name = "Float"
    _scalar_encoder = str
    _scalar_allowed_type = [float, int]
    scalar_type_hint = float


class IntScalarHandler(ScalarMixin, BaseHandler):
    """Handler for GraphQL Int scalar type."""

    _scalar_name = "Int"
    _scalar_encoder = str
    _scalar_allowed_type = int
    scalar_type_hint = int


class BooleanScalarHandler(ScalarMixin, BaseHandler):
    """Handler for GraphQL Boolean scalar type."""

    _scalar_name = "Boolean"
    _scalar_encoder = json.dumps
    _scalar_allowed_type = bool
    scalar_type_hint = bool


class StringScalarHandler(ScalarMixin, BaseHandler):
    """Handler for GraphQL String scalar type."""

    _scalar_name = "String"
    _scalar_encoder = json.dumps
    _scalar_allowed_type = str
    scalar_type_hint = str


class IDScalarHandler(ScalarMixin, BaseHandler):
    """Handler for GraphQL ID scalar type."""

    _scalar_name = "ID"
    _scalar_encoder = json.dumps
    _scalar_allowed_type = str
    scalar_type_hint = str



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from .base import BaseHandler
from .enum_handler import EnumHandler
from .input_object_handler import (
    InputObjectHandler,
    NamedInputObjectHandler,
)
from .list_handler import ListHandler
from .non_null_handler import NonNullHandler
from .object_handler import (
    NamedObjectHandler,
    ObjectHandler,
)
from .scalar_handler import (
    BooleanScalarHandler,
    FloatScalarHandler,
    IDScalarHandler,
    IntScalarHandler,
    StringScalarHandler,
)
from .union_handler import UnionHandler



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for `AsyncResult`."""
import logging
from typing import (
    List,
    Optional,
)

from graphql import GraphQLType
from tqdm.auto import tqdm

from .result_mixin import ResultMixin
from .utils import _get_function_name

LOGGER = logging.getLogger(__name__)


def _format_action(result: ResultMixin):
    """
    Returns a string representing the result's action (including name and action ID).
    """
    func_name = _get_function_name(result.action.name)
    action_id = result.action_id

    return f'{func_name} (action_id="{action_id}")'


class AsyncResult:
    """Wrapper around the result object which performs
    updating and progress reporting."""

    # statuses
    PENDING = "PENDING"
    STARTED = "STARTED"
    SUCCESS = "SUCCESS"

    def __init__(self, field_type: GraphQLType, result_object: ResultMixin):
        self._field_type = field_type
        self._result_object = result_object
        self._sent_messages = set()

    @property
    def result_object(self) -> ResultMixin:
        """Exposes the internal result object."""
        return self._result_object

    def update(self, env: "GraphQLEnvironment", data: dict, progress_bar: tqdm = None):
        """Updates the result object with the response data. Prints
        any progress updates to the progress bar."""

        env.update_data(self._field_type, self._result_object, data)

        # display progress message if required
        if progress_bar is not None:
            message = self._get_progress_message()

            if message and message not in self._sent_messages:
                progress_bar.write(message)
                self._sent_messages.add(message)

    def _get_progress_message(self) -> Optional[str]:
        """Returns the current progress message."""

        status = self._result_object.action.status
        action = _format_action(self._result_object)

        if status == self.PENDING:
            return (
                f"Your task {action} is currently in a queue waiting to be processed."
            )

        if status == self.STARTED:
            return (
                f"Your task {action} has started. You can use the `qctrl.get_result` method to "
                "retrieve previous results."
            )

        if status == self.SUCCESS:
            return f"Your task {action} has completed."

        return None


class AsyncResultCollection:
    """Represents a set of AsyncResult objects made in the one request."""

    def __init__(self, *async_results: AsyncResult):
        self._async_results = async_results
        self._current_progress = 0

    @property
    def is_completed(self) -> bool:
        """Checks if all AsyncResult objects are completed."""
        return all(  # pylint:disable=use-a-generator
            [
                async_result.result_object.is_completed
                for async_result in self._async_results
            ]
        )

    def update_progress(self, progress_bar: tqdm):
        """Updates the progress bar according to the
        current progress of the AsyncResult objects."""
        new_progress = [
            async_result.result_object.progress * 100
            for async_result in self._async_results
        ]
        new_progress = int(sum(new_progress) / len(new_progress))

        delta = new_progress - self._current_progress
        self._current_progress += delta
        progress_bar.update(delta)

    def get_action_ids(self) -> List[str]:
        """Returns the action IDs for the AsyncResult objects."""
        return [
            async_result.result_object.action_id for async_result in self._async_results
        ]

    def update_data(
        self, env: "GraphQLEnvironment", response_data: dict, progress_bar: tqdm
    ):
        """Updates the related AsyncResult objects with the response data."""

        # a single action will be returned as a dict
        if isinstance(response_data, dict):
            response_data = [response_data]

        if len(self._async_results) != len(response_data):
            raise RuntimeError(
                f"Unexpected responses ({len(response_data)}) for "
                f"results ({len(self._async_results)})"
            )

        for index, async_result in enumerate(self._async_results):
            _response_data = response_data[index]
            async_result.update(env, _response_data, progress_bar)

    def finalize(self):
        """Reports on final status."""

        # List of (identifier string, exception message) tuples for each failure.
        failures = []

        for async_result in self._async_results:

            LOGGER.info(
                "Action %s finished with status: %s",
                str(async_result.result_object.action_id),
                async_result.result_object.status,
            )

            if not async_result.result_object.is_successful:

                identifier = _format_action(async_result.result_object)

                if async_result.result_object.job_errors:
                    LOGGER.info(
                        "Action %s failed: %s",
                        async_result.result_object.action_id,
                        async_result.result_object.job_errors.exception,
                    )
                    exception = f"{async_result.result_object.job_errors.exception}"

                elif async_result.result_object.status == "FAILURE":
                    LOGGER.info(
                        "Action %s resulted with status 'FAILURE'("
                        "no extra details about the errors are "
                        "available).",
                        async_result.result_object.action_id,
                    )
                    exception = "Unknown failure"

                failures.append((identifier, exception))

        if not failures:
            return

        failure_message = "\n".join(": ".join(failure) for failure in failures)

        if len(self._async_results) == 1:
            raise RuntimeError(failure_message)

        raise RuntimeError(
            f"Action failures: {len(failures)} of {len(self._async_results)}\n"
            f"{failure_message}"
        )



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import datetime
from typing import Union

from gql import gql

from qctrl.queries.base import StaticQuery


class ActionExportQuery(StaticQuery):
    """
    Extracts data from the actions table.
    """

    query = gql(
        """
        query exportActions($limit: Int, $filterBy: ActionFilter) {
            actions(limit: $limit, filterBy: $filterBy){
                actions {
                    user {
                      username
                    }
                    modelId
                    name
                    status
                    errors {
                        exception
                        traceback
                    }
                    createdAt
                    updatedAt
                    terminatedAt
                    runtime
                }
                errors {
                    message
                }
            }
        }
    """
    )

    MAX_RECORDS = 50000

    def _get_variable_values(  # pylint:disable=arguments-differ
        self,
        start_date: Union[datetime.date, str],
        end_date: Union[datetime.date, str],
        ignore_test_users: bool = True,
    ):

        if isinstance(start_date, str):
            start_date = self.str_to_date(start_date)

        if isinstance(end_date, str):
            end_date = self.str_to_date(end_date)

        if start_date > end_date:
            raise ValueError(f"Invalid date range: {start_date} to {end_date}")

        filter_by = {
            "createdAt": {
                "range": [self.date_to_str(start_date), self.date_to_str(end_date)]
            },
            "ignoreTestUsers": ignore_test_users,
        }

        return {"limit": self.MAX_RECORDS, "filterBy": filter_by}

    def _format_response(self, response: dict, *_) -> dict:
        return response["actions"]["actions"]

    @staticmethod
    def str_to_date(date_str: str) -> datetime.date:
        """Helper function to convert a date string
        to a datetime.date object."""
        return datetime.datetime.strptime(date_str, "%Y-%m-%d").date()

    @staticmethod
    def date_to_str(date: datetime.date) -> str:
        """Helper function to convert a datetime.date
        object to a string."""
        return date.strftime("%Y-%m-%d")



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
import os
import sys
import webbrowser
from threading import Thread
from typing import (
    Dict,
    Optional,
    Union,
)

import click
import tenacity
from gql import (
    Client,
    gql,
)
from qctrlcommons.exceptions import QctrlException
from requests.exceptions import (
    BaseHTTPError,
    RequestException,
)

from .builders import (
    AsyncResult,
    build_and_bind_namespaces,
    create_client_auth,
    create_environment,
    create_gql_client,
)
from .constants import (
    BASE_URL,
    DEFAULT_API_ROOT,
    DEFAULT_OIDC_URL,
    INVALID_SUBSCRIPTION_ERROR,
    UNEXPECTED_TOKEN_ERROR,
)
from .graphs import Graph
from .parallel import ParallelExecutionCollector
from .queries import (
    ActivityMonitorQuery,
    GetMutationNameQuery,
    MeQuery,
)
from .utils import (
    PackageRegistry,
    _get_mutation_result_type,
)

LOGGER = logging.getLogger(__name__)


class Qctrl:
    """A mediator class. Used to authenticate with Q-CTRL and access Q-CTRL features.

    Creating an instance of this class requires authentication with Q-CTRL's API.

    The recommended method of authentication is through the interactive authentication
    method. This method can be invoked by simply calling `Qctrl()` without any arguments.
    This method will also create an authentication file that will be used for subsequent
    authentications when using the package.

    .. code-block:: python

      qctrl = Qctrl()

    You can also access toolkits provided by Boulder Opal through the corresponding namespaces
    of the Qctrl object.
    See the `reference <https://docs.q-ctrl.com/boulder-opal/references/qctrl/Toolkits.html>`_
    for details.

    .. deprecated:: 15.4.4
          Email and password based authentication has been removed in favour of
          url based authentication. To authenticate simply call `qctrl = Qctrl()`
          without any additional arguments.

    Parameters
    ----------
    email : str, optional
        The email address for a Q-CTRL account.
        [This field has been deprecated] (Default value = None)
    password : str, optional
        The password for a Q-CTRL account. [This field has been deprecated] (Default value = None)
    api_root : str, optional
        The URL of the Q-CTRL API. (Default value = None)
    oidc_url : str, optional
        The URL of the Q-CTRL OIDC. (Default value = None)
    env_namespace : str, optional
        The environment namespace of the Q-CTRL stack. (Default value = None)
    skip_version_check: bool, optional
        Option for disabling the version check. (Default value = False)
    client : gql.Client, optional
        A GraphQL client that provides access to a Q-CTRL GraphQL endpoint. You can pass
        this parameter to use Q-CTRL features provided by a non-standard Q-CTRL API
        implementation, for example, one running locally or in a private cloud. If you
        pass this parameter, do not pass `email`, `password`, or `api_root`.
        (Default value = None)

    Attributes
    ----------
    functions : :ref:`qctrl.dynamic.namespaces.FunctionNamespace`
    types : :ref:`qctrl.dynamic.types`

    Raises
    ------
    QctrlApiException
    """

    gql_api = None
    functions = None
    types = None

    def __init__(
        self,
        email: str = None,
        password: str = None,
        api_root: str = None,
        oidc_url: str = None,
        env_namespace: Optional[str] = None,
        skip_version_check: bool = False,
        client: Client = None,
    ):
        if not skip_version_check:
            self._check_version_thread()

        if email or password:
            error_message = (
                "Q-CTRL Python package authentication has been updated."
                " Please start a session without your credentials."
            )
            click.echo(error_message)
            sys.exit(1)

        if client is None:
            api_root = api_root or os.environ.get("QCTRL_API_HOST") or DEFAULT_API_ROOT
            oidc_url = oidc_url or os.environ.get("QCTRL_OIDC_URL") or DEFAULT_OIDC_URL

            env_namespace = env_namespace or os.environ.get("QCTRL_ENV_NAMESPACE")
            if env_namespace:
                api_root = BASE_URL.format(f"{env_namespace}.api")
                oidc_url = BASE_URL.format(f"{env_namespace}.id")

            self.gql_api = _build_client(api_root=api_root, oidc_url=oidc_url)
        else:
            if env_namespace or oidc_url or api_root:
                raise ValueError(
                    "If you pass a client, do not pass an env_namespace, oidc_url, or api_root."
                )
            self.gql_api = client

        self.gql_env = create_environment(self.gql_api)
        self.collector: Optional[ParallelExecutionCollector] = None
        self._build_and_bind_namespaces()

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=5),
        stop=tenacity.stop_after_attempt(3),
        retry=tenacity.retry_if_exception_type((RequestException, BaseHTTPError)),
    )
    def _build_and_bind_namespaces(self):
        """Builds the dynamic namespaces."""
        build_and_bind_namespaces(self)

    @staticmethod
    def create_graph() -> Graph:
        """
        Creates a graph object for representing remote computations.

        Returns
        -------
        Graph
            The new graph object.
        """
        return Graph()

    def activity_monitor(
        self,
        limit: int = 5,
        offset: int = 0,
        status: str = None,
        action_type: str = None,
    ) -> None:
        """Prints a list of previously run actions to the console
        and their statuses. Allows users to filter the amount of
        actions shown as well as provide an offset.

        Parameters
        ----------
        limit : int
            The number of previously ran actions to show.(Default is 5)
        offset : int
            Offset the list of actions by a certain amount.
        status : str
            The status of the action.
        action_type : str
            The action type.
        """

        query = ActivityMonitorQuery(self.gql_api)
        query(limit=limit, offset=offset, status=status, action_type=action_type)

    def get_result(
        self, action_id: Union[str, int]
    ) -> "qctrl.dynamic.types.CoreActionResult":
        """This function is used to return the results of a previously run function.
        You will be able to get the id of your action from the activity monitor.

        Parameters
        ----------
        action_id: str
            the id of the action which maps to an executed function.

        Returns
        -------
        qctrl.dynamic.types.CoreActionResult
            an instance of a class derived from a CoreActionResult.
        """
        action_id = str(action_id)

        query = GetMutationNameQuery(self.gql_api)
        mutation_name = query(action_id)

        field_type = _get_mutation_result_type(self.gql_api.schema, mutation_name)

        refresh_query = self.gql_env.build_refresh_query(field_type)
        data = refresh_query(action_id)

        result = self.gql_env.load_data(field_type, data)
        self.gql_env.wait_for_completion(refresh_query, AsyncResult(field_type, result))
        return result

    def parallel(self) -> ParallelExecutionCollector:
        """
        Context manager for executing multiple function calls in parallel.

        Any :ref:`functions <qctrl.dynamic.namespaces.FunctionNamespace>` that you call inside the
        context manager will be scheduled for execution when the context manager exits. For
        example:

        .. code-block:: python

          with qctrl.parallel():
              result_1 = qctrl.functions.calculate_optimization(...)
              result_2 = qctrl.functions.calculate_optimization(...)
              # The functions get executed when the context manager exits, so result_1 and result_2
              # do not have the results of the optimizations yet.

          # Once outside the context manager, the functions have been executed, so result_1 and
          # result_2 now contain the optimization results.

        Returns
        -------
        ParallelExecutionCollector
            The context manager that collects function calls to be executed in parallel.
        """
        return ParallelExecutionCollector(self)

    def is_collecting(self) -> bool:
        """Checks if the object is in collection mode.

        Returns
        -------
        bool
            True if in collection mode, False otherwise.
        """
        return bool(self.collector)

    def start_collection_mode(self, collector: ParallelExecutionCollector):
        """Starts collection mode. All function calls will be collected
        and executed when the collector object exits.

        Parameters
        ----------
        collector: ParallelExecutionCollector
            the collector object where function calls are stored.

        Raises
        ------
        RuntimeError
            unable to enter collection mode if already collecting.
        """
        if self.is_collecting():
            raise RuntimeError("unable to nest parallel collections")

        self.collector = collector

    def stop_collection_mode(self):
        """Stops collection mode. Function calls will be executed
        immediately.
        """
        self.collector = None

    def _run_gql_query(self, query: str, variable_values: Dict = None) -> Dict:
        """
        Runs a GQL query in a Python script.

        Parameters
        ----------
        query: str
            query string.
        variable_values: Dict
            Dictionary of input parameters. (Default value = None)

        Returns
        -------
        Dict
            gql response.

        Raises
        ------
        QctrlException
            if there's any root level errors
        """
        response = self.gql_api.execute(gql(query), variable_values)
        if response.get("errors"):
            raise QctrlException(response["errors"])
        return response

    @staticmethod
    def _check_version_thread():
        """
        Use another thread to check qctrl version.
        """
        for package in PackageRegistry:
            Thread(target=package.check_latest_version, daemon=True).start()


@tenacity.retry(
    wait=tenacity.wait_exponential(multiplier=1, min=1, max=5),
    stop=tenacity.stop_after_attempt(3),
    retry=tenacity.retry_if_exception_type((RequestException, BaseHTTPError)),
)
def _build_client(api_root: str, oidc_url: str) -> Client:
    """
    Builds an authenticated GraphQL client and validates its authentication.

    Parameters
    ----------
    api_root: str
        API root (Default value = None)
    oidc_url: str
        OIDC url (Default value = None)

    Returns
    -------
    GraphQLClient
        gql client.
    """

    try:
        auth = create_client_auth(oidc_url)
        client = create_gql_client(api_root, auth)
        MeQuery(client)()

    except QctrlException as error:
        LOGGER.info(error, exc_info=error)

        if "token" in str(error):
            click.echo(UNEXPECTED_TOKEN_ERROR, err=True)
            sys.exit(1)

        if "invalid subscription" in str(error):
            if "api." in api_root:
                webbrowser.open(api_root.replace("api.", "boulder."))

            click.echo(INVALID_SUBSCRIPTION_ERROR, err=True)
            sys.exit(1)

        sys.exit(error)

    return client


__all__ = ["Qctrl"]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring, too-few-public-methods

import re
import shutil
from functools import wraps
from textwrap import fill
from typing import (
    Callable,
    List,
    Optional,
)

import attr
import commonmark
import inflection

from qctrl.utils import (
    _convert_md_to_rst,
    _parse_to_rst,
)


def _get_title_underline(title: str, char: str = "-") -> str:
    """Returns the underline for the title of a docstring section.

    Parameters
    ----------
    title: str
        section title.
    char: str
        underline char. (Default value = "-")

    Returns
    -------
    str
        underline in str.
    """
    return "".join([char for _ in range(len(title))])


def docstring_section(title: str) -> Callable:
    """Builds a section of the docstring e.g. Parameters, Returns, etc.

    The decorated method should return a BaseDoc or list of
    BaseDoc objects.

    Parameters
    ----------
    title: str
        section title.

    Returns
    -------
    Callable
        decorator method
    """

    def customized_decorator(func):
        @wraps(func)
        def decorator(self, *args, **kwargs) -> Optional[str]:
            value = func(self, *args, **kwargs)

            if not value:
                return None

            lines = [title, _get_title_underline(title)]

            # single BaseDoc object
            if isinstance(value, BaseDoc):
                lines.append(str(value))

            # list of BaseDoc objects
            elif isinstance(value, (tuple, list)):
                for item in value:
                    if not isinstance(item, BaseDoc):
                        raise ValueError(f"unexpected list item: {item}")

                    lines.append(str(item))

            else:
                raise ValueError(f"unexpected function result: {value}")

            return "\n".join(lines)

        return property(decorator)

    return customized_decorator


class BaseDoc:
    """The base class for all docstring pieces.

    Subclasses should implement the __str__ method.
    """

    _indent_char = " "
    _indent_length = 4

    @property
    def _indent_str(self) -> str:
        return self._indent_char * self._indent_length

    def _indent(self, text: str) -> str:
        terminal_width = shutil.get_terminal_size().columns

        # In numpydoc standard, the length of docstring is up to 75.
        # Convert width to 75 if terminal_width is larger than 75
        # (and to 1 if terminal_width is reported as 0, otherwise
        # `fill` throws an error).
        width = max(1, min(75, terminal_width))

        # wrap the single line description into paragraph
        return fill(
            text,
            width=width,
            initial_indent=self._indent_str,
            subsequent_indent=self._indent_str,
        )

    def __str__(self):
        """To be implemented by the subclass."""
        raise NotImplementedError


class Docstring(BaseDoc):  # pylint:disable=too-few-public-methods
    """Base class for dynamically building docstrings consisting of a number of
    sections.

    Attributes
    ----------
    _sections: list of str
        list of ordered attributes corresponding to
        sections of the docstring.
    """

    _sections = []

    def _get_sections(self) -> List[str]:
        """Returns the list of sections for the docstring."""
        sections = []

        for _attr in self._sections:
            value = getattr(self, _attr)

            if value:
                sections.append(value)

        return sections

    def __str__(self):
        sections = self._get_sections()
        return "\n\n".join(sections)


@attr.s(auto_attribs=True)
class ClassAttribute(BaseDoc):
    """Describes a class attribute.

    Format:
    name: attr_type
        description

    Attributes
    ----------
    name: str
        The name of the attribute.
    attr_type: str
        The type of the attribute.
    description: str
        Verbose description of the attribute.
    optional: bool
        If the attribute is optional.
    """

    name: str
    attr_type: str = None
    description: Optional[str] = attr.ib(default=None, converter=_convert_md_to_rst)
    optional: bool = None

    def __str__(self):
        result = self.name

        if self.attr_type:
            result = f"{result}: {self.attr_type}"

        if self.optional:
            result = f"{result}, optional"

        if self.description:
            result += f"\n{self._indent(self.description)}"

        return result


@attr.s(auto_attribs=True)
class ClassDocstring(Docstring):
    """Constructs a doc string for a class."""

    description: Optional[str] = attr.ib(default=None, converter=_convert_md_to_rst)
    attrs: List[ClassAttribute] = attr.ib(default=attr.Factory(list))

    _sections = ["fdescription", "fattrs"]

    @property
    def fdescription(self):
        """Ensures summary is non-empty for formatting."""
        return self.description or " "

    @docstring_section("Attributes")
    def fattrs(self) -> List[ClassAttribute]:
        """Formatted Attributes section."""
        return self.attrs


@attr.s(auto_attribs=True)
class FunctionArgument(BaseDoc):
    """Describes an argument to a function.

    Format:
    name (arg_type, optional): description

    Attributes
    ----------
    name: str
        The name of the argument.
    arg_type: str
        The type of the argument.
    description: str
        Verbose description of the argument.
    optional: bool
        If the argument is optional.
    """

    name: str
    arg_type: str = None
    description: Optional[str] = attr.ib(default=None, converter=_convert_md_to_rst)
    optional: bool = None

    def __str__(self):
        result = self.name

        if self.arg_type:
            result = f"{result}: {self.arg_type}"

        if self.optional:
            result = f"{result}, optional"

        if self.description:
            result = f"{result}\n{self._indent(self.description)}"

        return result


@attr.s(auto_attribs=True)
class PossibleException(BaseDoc):
    """Describes a possible exception that a function could raise.

    Format:
    name
        description

    Attributes
    ----------
    name: str
        The class name of the exception.
    description: str
        Verbose description of the exception and why it would
        be raised.
    """

    name: str
    description: Optional[str] = attr.ib(default=None, converter=_convert_md_to_rst)

    def __str__(self):
        result = self.name

        if self.description:
            result = f"{result}\n{self._indent(self.description)}"

        return result


@attr.s(auto_attribs=True)
class Returns(BaseDoc):
    """Describes the object returned from a function.

    Format:
    return_type
        description

    Parameters
    ----------

    Returns
    -------

    Attributes
    ----------
    return_type: str
        cls type
    description: str
        Description of the returned object.
    """

    return_type: str
    description: Optional[str] = attr.ib(default=None, converter=_convert_md_to_rst)

    def __str__(self):
        result = self.return_type

        if self.description:
            # merge multiple lines to one line to avoid creating an item list
            merged_line_description = " ".join(self.description.splitlines())
            result = f"{result}\n{self._indent(merged_line_description)}"

        return result


@attr.s(auto_attribs=True)
class WarningsType(BaseDoc):
    """Describes the warnings section of NumpyDoc."""

    warnings: str

    def __str__(self):
        return self.warnings


@attr.s(auto_attribs=True)
class SeeAlsoType(BaseDoc):
    """Describes the see also section of NumpyDoc."""

    see_also: str

    def __str__(self):
        return self.see_also


@attr.s(auto_attribs=True)
class NotesType(BaseDoc):
    """Describes the notes section of NumpyDoc."""

    notes: str

    def __str__(self):
        return self.notes


@attr.s(auto_attribs=True)
class ReferencesType(BaseDoc):
    """Describes the references section of NumpyDoc."""

    references: str

    def __str__(self):
        """Replacing the references with reST specific directives."""
        reference_link_regex = r"\[([\d.]+)\]_"
        return re.sub(reference_link_regex, r".. [\1]", self.references)


@attr.s(auto_attribs=True)
class ExamplesType(BaseDoc):
    """Describes the examples section of NumpyDoc."""

    examples: str

    def __str__(self):
        return self.examples


@attr.s(auto_attribs=True)
class FunctionDocstring(Docstring):
    """Constructs a docstring for a function."""

    description: str = None
    params: List[FunctionArgument] = attr.ib(default=attr.Factory(list))
    returns: Returns = None
    raises: List[PossibleException] = attr.ib(default=attr.Factory(list))
    warnings: str = None
    see_also: str = None
    notes: str = None
    references: str = None
    examples: str = None

    _sections = [
        "description",
        "fparams",
        "freturns",
        "fraises",
        "fwarnings",
        "fsee_also",
        "fnotes",
        "freferences",
        "fexamples",
    ]

    @docstring_section("Parameters")
    def fparams(self) -> Optional[str]:
        """Formatted Args section."""
        return self.params

    @docstring_section("Returns")
    def freturns(self) -> Optional[str]:
        """Formatted Returns section."""
        return self.returns

    @docstring_section("Raises")
    def fraises(self) -> Optional[str]:
        """Formatted Raises section."""
        return self.raises

    @docstring_section("Warnings")
    def fwarnings(self) -> Optional[str]:
        """Formatted Warnings section."""
        return WarningsType(self.warnings) if self.warnings else None

    @docstring_section("See Also")
    def fsee_also(self) -> Optional[str]:
        """Formatted See Also section."""
        return SeeAlsoType(self.see_also) if self.see_also else None

    @docstring_section("Notes")
    def fnotes(self) -> Optional[str]:
        """Formatted Notes section."""
        return NotesType(self.notes) if self.notes else None

    @docstring_section("References")
    def freferences(self) -> Optional[str]:
        """Formatted Examples section."""
        return ReferencesType(self.references) if self.references else None

    @docstring_section("Examples")
    def fexamples(self) -> Optional[str]:
        """Formatted References section."""
        return ExamplesType(self.examples) if self.examples else None

    @classmethod
    def from_markdown(cls, markdown_text: str) -> "FunctionDocstring":
        """Generates an rst formatted FunctionDocstring from Markdown text.

        Parameters
        ----------
        markdown_text: str
            The markdown formatted docstring to process.

        Returns
        -------
        FunctionDocstring
            The FunctionDocstring class populated with the relevant sections.
        """
        _section_headings = (
            "description",
            "warnings",
            "see_also",
            "notes",
            "references",
            "examples",
        )
        docstring = cls()
        if markdown_text is None:
            return docstring

        # parse to ast
        parser = commonmark.Parser()
        ast = parser.parse(markdown_text)

        # start at the first child
        ast = ast.first_child

        current_text = ""

        if ast.t == "heading":
            current_heading = cls._parse_md_heading(ast.first_child.literal)
        else:
            current_heading = "description"
            current_text = _parse_to_rst(ast)

        event = ast.nxt
        sections = {}

        while event is not None:
            if event.t == "heading" and event.level == 1 and event.first_child.literal:
                sections[current_heading] = current_text
                heading = cls._parse_md_heading(event.first_child.literal)
                current_heading = heading
                current_text = ""
            else:  # it's a paragraph
                rst = _parse_to_rst(event) or ""
                current_text += rst

            event = event.nxt

        sections[current_heading] = current_text

        for key, value in sections.items():
            if key in _section_headings and value != "":
                setattr(docstring, key, value)

        return docstring

    @staticmethod
    def _parse_md_heading(heading: str) -> str:
        return inflection.underscore(inflection.parameterize(heading))


def add_deprecation_warning(object_name: str) -> str:
    """
    add deprecation warning for factories and models.

    Parameters
    ----------
    object_name: str
        name of the model/factory.

    Returns
    -------
    str
        formatted warning docstring.
    """
    deprecation_warning = f"""
Warning
-------
`{object_name}` will be deprecated soon. Please use
:class:`qctrl.dynamic.namespaces.TypeNamespace` instead.
"""
    return deprecation_warning



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring,too-many-ancestors
import json
from datetime import datetime
from typing import Any

import pytz
from dateutil.parser import parse
from qctrlcommons.serializers import DataTypeEncoder

from qctrl.builders.graphql_utils.handlers.base import BaseHandler
from qctrl.builders.graphql_utils.handlers.mixins import ScalarMixin
from qctrl.graphs import Graph


def _encode_custom_scalar(value: Any) -> Any:
    """Encodes a custom scalar value."""
    result = json.dumps(value, cls=DataTypeEncoder)
    return json.dumps(result)


def _decode_datetime_scalar(value: str) -> datetime:
    """
    Decode a datetime scalar value.
    """
    return parse(value)


def _decode_unix_scalar(value: str) -> datetime:
    """
    Decode a unixtime scalar value.
    """
    return datetime.fromtimestamp(int(value)).astimezone(pytz.utc)


class DateTimeScalarHandler(ScalarMixin, BaseHandler):
    """Custom scalar handler for DateTime."""

    _scalar_name = "DateTime"
    _scalar_allowed_type = datetime
    _scalar_decoder = _decode_datetime_scalar
    scalar_type_hint = datetime


class GraphScalarHandler(ScalarMixin, BaseHandler):
    """Custom scalar handler for Graph."""

    _scalar_name = "Graph"
    _scalar_encoder = _encode_custom_scalar
    _scalar_allowed_type = Graph
    scalar_type_hint = Graph


class JsonDictScalarHandler(ScalarMixin, BaseHandler):
    """Custom scalar handler for JsonDict."""

    _scalar_name = "JsonDict"
    _scalar_encoder = _encode_custom_scalar
    _scalar_allowed_type = dict
    scalar_type_hint = dict


class UnixTimeScalarHandler(ScalarMixin, BaseHandler):
    """Custom scalar handler for UnixTime."""

    _scalar_name = "UnixTime"
    _scalar_allowed_type = datetime
    _scalar_decoder = _decode_unix_scalar
    scalar_type_hint = datetime



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Redirect listener.
"""
import os
import socket
import webbrowser
from http.server import (
    BaseHTTPRequestHandler,
    HTTPServer,
)

BASE_REDIRECT_URL = "https://accounts.q-ctrl.com/auth"
REDIRECT_SUCCESS_URL = f"{BASE_REDIRECT_URL}/success?messageid=pythonpackage"
REDIRECT_ERROR_URL = f"{BASE_REDIRECT_URL}/error?messageId=pythonPackage"


def check_if_network_port_is_available(port: int):
    """
    Check if a network port is available.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        return sock.connect_ex(("", port)) != 0


def get_free_network_port(range_from: int = 8000, range_to: int = 9000) -> int:
    """
    Get a free network port.
    """
    for port in range(range_from, range_to):
        if check_if_network_port_is_available(port):
            return port
    raise RuntimeError("No free network port found")


def complete_login(port, url, callback):
    """
    Creates a simple web server that will open a browser to the login page and
    listen for the redirect callback.
    """

    class RequestHandler(BaseHTTPRequestHandler):
        """
        Basic HTTP server that listen to a single request and redirects the path
        to the context callback function.
        """

        def log_request(self, *args, **kwargs) -> None:
            """
            Only log the request if `DEBUG_OIDC` is set.
            """
            if os.environ.get("DEBUG_OIDC"):
                super().log_request(*args, **kwargs)

        def do_GET(self):  # pylint: disable=invalid-name
            """
            Handle GET requests.
            """
            try:
                callback(self.path)

            except:  # pylint: disable=bare-except
                self.send_response(307)
                self.send_header("Location", REDIRECT_ERROR_URL)
                self.end_headers()

            else:
                self.send_response(307)
                self.send_header("Location", REDIRECT_SUCCESS_URL)
                self.end_headers()

    server_address = ("", port)
    try:
        httpd = HTTPServer(server_address, RequestHandler)
    except OSError as exc:
        raise RuntimeError(
            "Could not start HTTP server, port is already in use"
        ) from exc
    else:
        print(
            "The URL above should be automatically opened in your default web browser."
        )
        print("(Please copy and paste in case it doesn't open automatically)")
        print("")
        print("Authenticate your credentials in your browser...")
        print("")
        webbrowser.open(url)
        httpd.handle_request()



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring,too-many-ancestors
from typing import (
    Any,
    Dict,
)

import inflection
import numpy as np
from qctrlcommons.serializers import write_numpy_array

from qctrl.builders.graphql_utils.handlers import NamedInputObjectHandler
from qctrl.builders.type_registry import TypeOverrideMixin


class ComplexArrayInputHandler(NamedInputObjectHandler, TypeOverrideMixin):
    """Custom handler for ComplexArrayInput."""

    _type_hint = np.ndarray
    _input_object_name = "ComplexArrayInput"

    def get_input_object_values(self, data: Dict[str, Any]):
        serialized_data = write_numpy_array(data)
        return {
            inflection.camelize(key, False): value
            for key, value in serialized_data.items()
        }



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging

from graphql import GraphQLField
from qctrlcommons.node.registry import (
    NODE_REGISTRY,
    TYPE_REGISTRY,
)

from qctrl import toolkit
from qctrl.utils import PackageRegistry

from .async_result import AsyncResult
from .client_builder import (
    create_client_auth,
    create_gql_client,
)
from .custom_environment import create_environment
from .doc import (
    FunctionArgument,
    FunctionDocstring,
)
from .functions import (
    build_function,
    build_signature,
)
from .namespaces import (
    BaseNamespace,
    ToolkitCategory,
    build_and_bind_toolkit,
    create_function_namespace,
    create_type_namespace,
)
from .result_mixin import ResultMixin
from .type_registry import TypeRegistry

LOGGER = logging.getLogger(__name__)


def _is_valid_function_mutation(name: str) -> bool:
    """Checks if the mutation field corresponds to a valid function which
    should be available to the user. Mutations that should be exposed have a
    single argument named `input`.

    Parameters
    ----------
    name: GraphQLField
        The mutation field to check.

    Returns
    -------
    bool
        True if field is a valid mutation to be exposed to a user.
    """
    return name.startswith("core__")


def build_and_bind_namespaces(qctrl: "Qctrl") -> None:
    """Builds features, organizes them into namespaces and binds the namespace to the
    qctrl object.

    Parameters
    ----------
    qctrl : Qctrl
        The Qctrl object.
    """
    function_namespace = create_function_namespace()
    type_namespace = create_type_namespace()

    # build namespace for functions and types for GQL
    mutation_type = qctrl.gql_api.schema.mutation_type

    for name, field in mutation_type.fields.items():
        if _is_valid_function_mutation(name):
            func = build_function(qctrl, name, field)
            function_namespace.extend_functions(func)

    type_namespace.add_registry(qctrl.gql_env.type_registry)

    setattr(qctrl, "functions", function_namespace)
    setattr(qctrl, "types", type_namespace)
    build_and_bind_toolkit(qctrl, ToolkitCategory.FUNCTIONS, toolkit)


__all__ = ["build_and_bind_namespaces"]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from typing import (
    Callable,
    Union,
)

import forge
from gql.transport.exceptions import TransportQueryError
from graphql import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputObjectType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLType,
    is_required_input_field,
)
from qctrlcommons.exceptions import QctrlGqlException
from qctrlcommons.utils import get_gql_error_dict

from qctrl.queries import StartActionQuery
from qctrl.utils import _is_undefined

from .async_result import AsyncResult
from .doc import (
    FunctionArgument,
    FunctionDocstring,
    Returns,
)
from .graphql_utils import GraphQLEnvironment
from .result_mixin import ResultMixin
from .utils import _get_function_name

LOGGER = logging.getLogger(__name__)


def build_function(
    qctrl: "Qctrl", mutation_name: str, mutation_field: GraphQLField
) -> Callable:
    """Returns a function which can be used to execute a GraphQL mutation.

    Parameters
    ----------
    qctrl: Qctrl
        The Qctrl object.

    mutation_name: str
        The name of the mutation.

    mutation_field: GraphQLField
        The corresponding mutation field.

    Returns
    -------
    func : Callable
        A callable function.
    """

    input_type = _get_mutation_input_type(mutation_field)

    def get_query(**kwargs) -> StartActionQuery:
        return qctrl.gql_env.build_mutation_query(
            mutation_name, input_type, mutation_field.type, kwargs
        )

    def func(**kwargs):
        disable_result_cache = kwargs.pop("disable_result_cache", False)
        query = get_query(**kwargs)

        LOGGER.debug("query:%s", query)

        if qctrl.is_collecting():
            result = query.create_result_object({"action": {}})
            qctrl.collector.add(query, AsyncResult(mutation_field.type, result))
            return result

        try:
            response = query(disable_result_cache=disable_result_cache)
        except TransportQueryError as exc:
            LOGGER.debug("query execution failed: %s", exc)
            error_dict = get_gql_error_dict(exc)
            raise QctrlGqlException(error_dict["message"]) from exc

        LOGGER.debug("response:%s", response)

        if mutation_name not in response:
            LOGGER.error("response:%s", response)
            raise ValueError(f"response missing mutation name: {mutation_name}")

        result = query.create_result_object(response[mutation_name])

        LOGGER.info(
            "Remote function %s called - Action ID: %s", result.name, result.action_id
        )
        refresh_query = query.get_refresh_query()
        qctrl.gql_env.wait_for_completion(
            refresh_query, AsyncResult(mutation_field.type, result)
        )

        return result

    # set function name
    func.__name__ = _get_function_name(mutation_name)

    # build signature and sign function
    signature = build_signature(qctrl.gql_env, input_type, mutation_field.type)

    func = signature(func)

    # build docstring
    docstring = FunctionDocstring.from_markdown(mutation_field.description)
    return_type = forge.fsignature(func).return_annotation
    output_field_type = _get_mutation_output_type(mutation_field.type)
    docstring.returns = Returns(
        return_type=f"{return_type.__module__}.{return_type.__name__}",
        description=output_field_type.description,
    )

    for field_name, field in input_type.fields.items():
        docstring.params.append(
            FunctionArgument(
                qctrl.gql_env.field_to_attr(field_name),
                arg_type=qctrl.gql_env.type_registry.get_doc_field_type(field.type),
                optional=not is_required_input_field(field),
                description=field.description,
            )
        )

    # set function docstring
    func.__doc__ = str(docstring)

    func.request = get_query

    return func


def build_signature(
    env: GraphQLEnvironment, input_type: GraphQLType, return_type: GraphQLType = None
):
    """
    Returns a function signature based on the input fields. All fields
    are added as keyword args.
    """

    args = []

    for field_name, field in input_type.fields.items():
        default_value = field.default_value

        # if no default value
        if _is_undefined(default_value):

            # use special flag for required field
            if isinstance(field.type, GraphQLNonNull):
                default_value = forge.FParameter.empty

            # otherwise, use None
            else:
                default_value = None

        # build arg and add to list
        args.append(
            forge.kwarg(
                env.field_to_attr(field_name),
                type=_get_type_hint(env, field.type),
                default=default_value,
            )
        )

    LOGGER.debug("signature args:%s", args)

    # build signature
    signature = forge.sign(*args, **forge.kwargs)

    if return_type:
        returns = _get_type_hint(env, return_type, base=ResultMixin)

        if returns is not forge.FParameter.empty:
            signature = forge.compose(signature, forge.returns(returns))

    return signature


def _get_mutation_input_type(field: GraphQLField) -> GraphQLType:
    """Given a mutation field, return the input
    argument type. Every mutation is expected to
    have exactly one argument, named `input`.

    Parameters
    ----------
    field: GraphQLField
        The mutation field


    Returns
    -------
    GraphQLType
        type of input arguments

    Raises
    ------
    KeyError
        if mutation field does not have `input` arg
    ValueError
        if argument is not type of GraphQLArgument.
    TypeError
        if the input type is not GraphQLNonNull or GraphQLInputObjectType.
    """
    try:
        input_arg = field.args["input"]
    except KeyError as error:
        raise KeyError(f"mutation field does not have input arg: {field}") from error

    if not isinstance(input_arg, GraphQLArgument):
        raise ValueError(f"unexpected input arg type: {input_arg}")

    _type = input_arg.type

    if not isinstance(_type, GraphQLNonNull):
        raise TypeError(f"expected non-null type; got: {_type}")

    _type = _type.of_type

    if not isinstance(_type, GraphQLInputObjectType):
        raise TypeError(f"expected input object type; got: {_type}")

    return _type


def _get_mutation_output_type(field_type: GraphQLType) -> GraphQLType:
    """
    Given a mutation field, return the output
    argument type. Every mutation is expected to
    have exactly one NonNull output type.

    Parameters
    ----------
    field_type: GraphQLType
        the type of the mutation field

    Returns
    -------
    GraphQLType
        the final output type.

    Raises
    ------
    TypeError
        if it is not GraphQLObjectType.
    """
    if isinstance(field_type, (GraphQLNonNull, GraphQLList)):
        return _get_mutation_output_type(field_type.of_type)

    if not isinstance(field_type, GraphQLObjectType):
        raise TypeError(f"expected object type; got: {field_type}")
    return field_type


def _get_type_hint(
    env: GraphQLEnvironment, field_type: GraphQLType, **build_opts
) -> Union[type, forge.FParameter.empty]:
    """
    Returns a type hint for the GraphQL field type. If a type registry
    is being used, `build_opts` can be used.

    for configuration if a new type is being created - see
    `TypeRegistry.get_type_hint`.
    If no type hint can be determined, forge.FParameter.empty
    is returned.

    Parameters
    ----------
    field_type: GraphQLType

    build_opts


    Returns
    -------
    Union[type, forge.FParameter.empty]
        type hint.
    """
    type_hint = env.type_registry.get_type_hint(field_type, **build_opts)
    return type_hint or forge.FParameter.empty



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    Optional,
)

from graphql import (
    GraphQLScalarType,
    GraphQLType,
)

from qctrl.utils import abstract_property


class BaseHandlerMixin(ABC):
    """
    Base Handler mixing for query or mutation types.
    """

    @staticmethod
    @abstractmethod
    def _can_handle(handler, field_type: GraphQLType) -> bool:
        """
        Private static method to be overridden by the mixin definition class.
        """

        raise NotImplementedError

    @classmethod
    def can_handle(cls, handler, field_type: GraphQLType) -> bool:
        """
        Checks if the handler is allowed for the given field type using static
        methods to allow dynamic checking.
        """

        return isinstance(handler, cls) and cls._can_handle(handler, field_type)


class PayloadMixin(BaseHandlerMixin):
    """Handler mixin for types that format a payload to
    be sent as a query/mutation request.
    """

    def can_handle_format_payload(self, field_type: GraphQLType) -> bool:
        """Checks if the handler is allowed for the given
        field type.
        """
        return self._is_expected_type(field_type)

    def format_payload(self, field_type: GraphQLType, data: Any) -> str:
        """Public method for formatting the given payload
        data for use in the query/mutation.
        """
        return self._format_payload(field_type, data)

    @abstractmethod
    def _format_payload(self, field_type: GraphQLType, data: Any) -> str:
        """Private method to be overridden for formatting
        the payload data.
        """
        raise NotImplementedError

    @staticmethod
    def _can_handle(handler, field_type: GraphQLType) -> bool:
        return handler.can_handle_format_payload(field_type)


class SelectionMixin(BaseHandlerMixin):
    """Handler mixin for types that are used for the selection
    part of a query or mutation selection.
    """

    def can_handle_get_selection(self, field_type: GraphQLType) -> bool:
        """Checks if the handler is allowed for the given
        field type.
        """
        return self._is_expected_type(field_type)

    def get_selection(
        self, field_type: GraphQLType, overrides: Optional[Dict[str, str]] = None
    ) -> str:
        """Public method for returning the selection query
        for the given field type. By default, all nested fields
        will be included. Overrides can be provided to restrict
        which nested fields are selected.
        """
        return self._get_selection(field_type, overrides)

    @abstractmethod
    def _get_selection(
        self, field_type: GraphQLType, overrides: Optional[Dict[str, str]]
    ) -> str:
        """Private method to be overridden for returning the
        query selection.
        """
        raise NotImplementedError

    @staticmethod
    def _can_handle(handler, field_type: GraphQLType) -> bool:
        return handler.can_handle_get_selection(field_type)


class DataMixin(BaseHandlerMixin):
    """Handler mixin for types that consume data returned by a
    query or mutation.
    """

    def can_handle_data(self, field_type: GraphQLType) -> bool:
        """Checks if the handler is allowed for the given
        field type.
        """
        return self._is_expected_type(field_type)

    def load_data(self, field_type: GraphQLType, data: Optional[Dict[str, Any]]) -> Any:
        """Public method for loading the given data according to
        the field type. If the data is None, no formatting is
        performed.
        """
        if data is None:
            return data

        return self._load_data(field_type, data)

    @abstractmethod
    def _load_data(self, field_type: GraphQLType, data: Dict[str, Any]) -> Any:
        """Private method to be overridden for loading the data."""
        raise NotImplementedError

    @staticmethod
    def _can_handle(handler, field_type: GraphQLType) -> bool:
        return handler.can_handle_data(field_type)

    def update_data(
        self, field_type: GraphQLType, obj: Any, data: Optional[Dict[str, Any]]
    ) -> Any:
        """Public method for updating an object with the given
        data according to the field type. If the data is None,
        no formatting is performed.
        """
        if data is None:
            return data

        return self._update_data(field_type, obj, data)

    @abstractmethod
    def _update_data(
        self, field_type: GraphQLType, obj: Any, data: Dict[str, Any]
    ) -> Any:
        """Private method to be overridden for updating the object
        with the data.
        """
        raise NotImplementedError


class ScalarMixin(PayloadMixin, SelectionMixin, DataMixin):
    """Handler mixin for all GraphQLScalar types."""

    _field_type = GraphQLScalarType
    _scalar_name = None
    _scalar_encoder = None
    _scalar_decoder = None
    _scalar_allowed_type = None
    scalar_type_hint = None

    @abstract_property
    def scalar_name(self):  # pylint:disable=missing-function-docstring
        return self._scalar_name

    @abstract_property
    def scalar_encoder(self):  # pylint:disable=missing-function-docstring
        return self.__class__._scalar_encoder  # pylint:disable=protected-access

    @property
    def scalar_decoder(self):  # pylint:disable=missing-function-docstring
        return self.__class__._scalar_decoder  # pylint:disable=protected-access

    @abstract_property
    def scalar_allowed_type(self):  # pylint:disable=missing-function-docstring
        return self._scalar_allowed_type

    def _is_expected_scalar_name(self, field_type: GraphQLType) -> bool:
        """Checks if field type name is expected
        for this handler.
        """
        return (
            field_type.name  # pylint:disable=comparison-with-callable
            == self.scalar_name
        )

    def _check_type(self, data: Any) -> bool:
        """Checks that the data type is expected."""
        allowed = self.scalar_allowed_type

        if not isinstance(allowed, list):
            allowed = [allowed]

        # pylint: disable=isinstance-second-argument-not-valid-type
        if not any(isinstance(data, _type) for _type in allowed):
            raise TypeError(
                f"scalar {self.scalar_name} expected object of "
                f"type: {self.scalar_allowed_type}; got {type(data)} ({data})"
            )

    # overridden methods

    def can_handle_format_payload(self, field_type):
        return super().can_handle_format_payload(
            field_type
        ) and self._is_expected_scalar_name(field_type)

    def _format_payload(self, field_type, data):
        self._check_type(data)
        return self.scalar_encoder(data)  # pylint:disable=too-many-function-args

    def can_handle_get_selection(self, field_type):
        return super().can_handle_get_selection(
            field_type
        ) and self._is_expected_scalar_name(field_type)

    def _get_selection(self, field_type, overrides=None):
        return None

    def can_handle_data(self, field_type):
        return super().can_handle_data(field_type) and self._is_expected_scalar_name(
            field_type
        )

    def _load_data(self, field_type, data):
        if self.scalar_decoder:
            data = self.scalar_decoder(data)  # pylint:disable=not-callable

        self._check_type(data)
        return data

    def _update_data(self, field_type, obj, data):
        # refresh query data is already decoded, so we can return the data directly.
        self._check_type(data)
        return data



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Tests for qctrlauth.redirect_listener module.
"""

import socket
from unittest.mock import call

import pytest

from qctrl.qctrlauth.redirect_listener import (
    check_if_network_port_is_available,
    complete_login,
    get_free_network_port,
)


def test_check_if_network_port_is_available():
    port = get_free_network_port()
    assert check_if_network_port_is_available(port) is True, "Port should be available"

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(("", port))
        sock.listen(1)
        assert (
            check_if_network_port_is_available(port) is False
        ), "Port shouldn't be available"
        sock.close()


def test_get_free_network_port():
    with pytest.raises(RuntimeError, match="No free network port found"):
        get_free_network_port(0, 0)


def test_complete_login(mocker):
    mocked_open = mocker.patch("webbrowser.open")
    mocked_handle = mocker.patch("http.server.HTTPServer.handle_request")

    port = get_free_network_port()
    complete_login(port, "http://test", lambda: None)

    assert mocked_open.called is True
    mocked_open.assert_has_calls([call("http://test")])
    assert mocked_handle.called is True


def test_complete_login_port_used():
    port = get_free_network_port()
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(("", port))
        sock.listen(1)
        with pytest.raises(RuntimeError, match="port is already in use"):
            complete_login(port, "http://test", lambda: None)
        sock.close()



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Q-CTRL OIDC Session.
"""

import json
import logging
import os
import time
from pathlib import Path
from typing import (
    Callable,
    Optional,
)

from jwt.api_jwt import decode
from oauthlib.oauth2 import WebApplicationClient
from oauthlib.oauth2.rfc6749.errors import (
    InvalidGrantError,
    TokenExpiredError,
)
from qctrlcommons.exceptions import QctrlException
from qctrlcommons.utils import get_nested_value
from requests_oauthlib import OAuth2Session

from .redirect_listener import (
    complete_login,
    get_free_network_port,
)

LOGGER = logging.getLogger(__name__)

# OAUTHLIB_INSECURE_TRANSPORT=1 disables the requirement for HTTPS for the
# localhost redirect server and allows "insecure" (HTTP) requests to our OIDC
# server as a side effect.
# As the servers will always validate tokens from clients against our trusted CA
# and our public services only accept HTTPS, it is safe to use this at client side.
os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"


class QctrlApplicationClient(WebApplicationClient):
    """
    QctrlApplicationClient is a subclass of WebApplicationClient that
    implements the verify_token helper method.
    """

    REFRESH_THRESHOLD = 60  # seconds

    def verify_token(self):
        """
        Verifies if the token is still valid or is expiring soon.
        """
        if not (self.access_token or self.token.get("access_token")):
            raise ValueError("Missing access token.")

        if self._expires_at and self._expires_at - time.time() < self.REFRESH_THRESHOLD:
            raise TokenExpiredError()


class QctrlOAuth2Session(OAuth2Session):
    """
    QctrlOAuth2Session is a subclass of OAuth2Session.

    This class implements some helper authentication related methods and overrides
    the `access_token` property to automatically refresh the token if it is expired.
    """

    _client: QctrlApplicationClient

    def __init__(
        self,
        client_id=None,
        client=None,
        base_url=None,
        scope=None,
        redirect_uri=None,
        redirect_uri_port=None,
        auto_refresh_kwargs=None,
        session_file_path: Path = None,
        token: Optional[dict] = None,
        token_updater: Optional[Callable] = None,
        **kwargs,
    ):
        """
        Initializes the QctrlOAuth2Session.
        """

        if redirect_uri and not redirect_uri_port:
            raise ValueError("redirect_uri_port is required when redirect_uri is set")

        if redirect_uri_port and not redirect_uri:
            raise ValueError("redirect_uri is required when redirect_uri_port is set")

        self.session_file_path = session_file_path
        self.redirect_uri_port = redirect_uri_port or get_free_network_port()
        redirect_uri = redirect_uri or f"http://localhost:{self.redirect_uri_port}"

        super().__init__(
            client_id=client_id,
            client=client or QctrlApplicationClient(client_id, redirect_uri),
            scope=scope or ["openid", "profile", "email", "offline_access"],
            auto_refresh_kwargs=auto_refresh_kwargs or {"client_id": client_id},
            redirect_uri=redirect_uri,
            token=token or self._token_loader,
            token_updater=token_updater or self._token_updater,
            **kwargs,
        )

        if base_url:
            if not base_url.endswith("/"):
                base_url += "/"

            oidc_base_url = f"{base_url}auth/realms/q-ctrl/protocol/openid-connect/"
            self.auth_url = f"{oidc_base_url}auth"
            self.token_url = f"{oidc_base_url}token"
            self.auto_refresh_url = f"{oidc_base_url}token"
            self.user_info_url = f"{oidc_base_url}userinfo"

    def _token_updater(self, token: dict):
        """
        Saves the token to the file.
        """
        if not self.session_file_path:
            return

        try:
            self.session_file_path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
            self.session_file_path.touch(mode=0o600, exist_ok=True)
            self.session_file_path.write_text(
                json.dumps(token, indent=2), encoding="utf-8"
            )
        except IOError as exc:
            LOGGER.error("%s", exc, exc_info=True)
            raise QctrlException("incorrect permissions for credentials file") from exc

    @property
    def _token_loader(self):
        """
        Loads the token from the file.
        """
        if not self.session_file_path:
            return {}

        try:
            return json.load(open(self.session_file_path, "r", encoding="utf-8"))
        except FileNotFoundError:
            return {}
        except IsADirectoryError as exc:
            raise QctrlException("credentials file cannot be a directory") from exc

    @property
    def access_token(self):
        try:
            self._client.verify_token()
            return self._client.access_token

        except TokenExpiredError:
            if self.auto_refresh_url:
                try:
                    self.token = self.refresh_token(self.auto_refresh_url)
                except Warning:
                    self.authenticate()

                if self.token_updater:
                    self.token_updater(self.token)
                return self._client.access_token
            raise

    def authenticate(self):
        """
        Generates the authorization url, opens it in the browser and starts the
        local server to listen for the redirect.
        """
        authorization_url, _ = self.authorization_url(self.auth_url)
        print("")
        print("Authentication URL:")
        print("")
        print(authorization_url)
        print("")
        complete_login(
            self.redirect_uri_port,
            authorization_url,
            self._fetch_token_from_authorization_response,
        )
        print("Successful authentication!")

    def authenticate_if_needed(self):
        """
        Verify if the `access_token` is still valid or can be refreshed, if not
        starts the `authenticate` flow.
        """
        try:
            if self.access_token and self.user_info:
                return
            self.authenticate()
        except (InvalidGrantError, ValueError):
            self.authenticate()

    @property
    def user_info(self):
        """
        Returns the user info from the OIDC server.
        """
        response = self.get(self.user_info_url)
        response.raise_for_status()
        return response.json()

    def _fetch_token_from_authorization_response(self, authorization_response):
        """
        Fetch token from authorization response and save it if `token_updater`
        is present.
        """
        print("Finalising authentication...")
        token = self.fetch_token(
            self.token_url, authorization_response=authorization_response
        )
        if self.token_updater:
            self.token_updater(token)

    def has_role(self, role) -> bool:
        """
        Returns True if the user has the given role for the client.
        """
        payload = decode(self.access_token, options={"verify_signature": False})
        client_roles = get_nested_value(
            payload, f"resource_access.{self.client_id}.roles", []
        )
        return role in client_roles

    def invalidate_access_token(self):
        """
        Invalidates the access token, forcing a refresh on next access.
        """
        if not self.token:
            return

        expired_access_token = self.token
        expired_access_token["expires_at"] = -1
        expired_access_token["expires_in"] = -1
        self.token_updater(expired_access_token)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qctrl',
 'qctrl.builders',
 'qctrl.builders.custom_handlers',
 'qctrl.builders.graphql_utils',
 'qctrl.builders.graphql_utils.handlers',
 'qctrl.qctrlauth',
 'qctrl.qctrlauth.tests',
 'qctrl.queries']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0,<20.0.0',
 'cached-property>=1.5.1,<2.0.0',
 'click>=8.0,<9.0',
 'commonmark>=0.9.1,<0.10.0',
 'gql[requests]==3.0.0rc0',
 'inflection>=0.5.1,<0.6.0',
 'numpy>=1.21.5,<2.0.0',
 'packaging>=20.9,<21.0',
 'pyjwt>=2.4.0,<3.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-forge>=18.6.0,<19.0.0',
 'pytz>=2020.1,<2021.0',
 'qctrl-commons>=16.4.0,<17.0.0',
 'qctrl-toolkit>=0.4.0,<0.5.0',
 'requests-oauthlib>=1.3.0,<2.0.0',
 'requests>=2.27.1,<2.28.0',
 'rich>=11.2.0,<12.0.0',
 'tenacity>=6.0.0,<7.0.0',
 'toml>=0.10.0,<0.11.0',
 'tqdm>=4.63.0',
 'typing-utils>=0.1.0,<0.2.0',
 'urllib3>=1.26.8,<1.27.0']

entry_points = \
{'console_scripts': ['qctrl = qctrl.scripts:main']}

setup_kwargs = {
    'name': 'qctrl',
    'version': '18.5.0',
    'description': 'Q-CTRL Python',
    'long_description': "# Q-CTRL Python\n\nThe Q-CTRL Python Package provides an intuitive and convenient Python interface\nto Q-CTRL's quantum control solutions for customers of Q-CTRL. To use the Q-CTRL\nPython Package, you will need a\n[Boulder Opal](https://q-ctrl.com/products/boulder-opal/) account.\n",
    'author': 'Q-CTRL',
    'author_email': 'support@q-ctrl.com',
    'maintainer': 'Q-CTRL',
    'maintainer_email': 'support@q-ctrl.com',
    'url': 'https://q-ctrl.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from abc import (
    ABC,
    abstractmethod,
)


class BaseTypeRegistry(ABC):  # pylint:disable=too-few-public-methods
    """Base class for a type registry which can be used
    for loading and updating data from a GraphQL response
    into objects.
    """

    def __init__(self, env: "GraphQLEnvironment"):
        self.env = env

    @abstractmethod
    def get_type(self, type_name: str) -> type:
        """Returns the corresponding class for the
        GraphQL field type name. To be overridden by
        the subclass.
        """
        raise NotImplementedError



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from functools import lru_cache
from typing import (
    Any,
    List,
    Union,
)

import inflection
from gql import Client
from graphql import GraphQLType
from tqdm.auto import tqdm

from ..queries import (
    MultiRefreshActionQuery,
    RefreshActionQuery,
    StartActionQuery,
)
from .async_result import (
    AsyncResult,
    AsyncResultCollection,
)
from .custom_handlers import (
    ComplexArrayEntryHandler,
    ComplexArrayHandler,
    ComplexArrayInputHandler,
    ComplexDenseArrayHandler,
    ComplexNumberHandler,
    ComplexNumberInputHandler,
    ComplexSparseArrayHandler,
    DateTimeScalarHandler,
    GraphScalarHandler,
    JsonDictScalarHandler,
    UnixTimeScalarHandler,
)
from .graphql_utils import (
    BaseHandler,
    GraphQLEnvironment,
)
from .type_registry import TypeRegistry
from .wait import Wait

LOGGER = logging.getLogger(__name__)


class QctrlGraphQLEnvironment(GraphQLEnvironment):
    """
    Custom GraphQLEnvironment class.
    """

    _action_selection_override = {
        "action": r"{ modelId name status errors { exception } progress }"
    }

    def __init__(
        self,
        gql_client: Client,
        type_registry_cls: type = None,
        custom_handlers: List[BaseHandler] = None,
    ):
        self._gql_client = gql_client
        super().__init__(type_registry_cls, custom_handlers)

    @lru_cache(maxsize=128)
    def field_to_attr(self, field_name: str) -> str:
        """Uses inflection to convert a GraphQL
        field name to an attribute name.

        Parameters
        ----------
        field_name : str
            The name of the GraphQL field.

        Returns
        -------
        str
            attribute name
        """
        return inflection.underscore(field_name)

    def build_mutation_query(
        self,
        mutation_name: str,
        input_type: GraphQLType,
        result_type: GraphQLType,
        data: Any,
    ) -> StartActionQuery:
        """Returns the corresponding StartActionQuery object."""
        query = StartActionQuery(
            self._gql_client, self, mutation_name, input_type, result_type, data
        )

        return query

    def wait_for_completion(
        self,
        refresh_query: Union[RefreshActionQuery, MultiRefreshActionQuery],
        *async_results: AsyncResult
    ):
        """Waits until the corresponding actions are completed on the server."""
        promise = AsyncResultCollection(*async_results)
        if promise.is_completed:
            promise.finalize()
            return

        _wait = Wait()

        with tqdm(total=100, leave=False) as progress_bar:

            while not promise.is_completed:

                promise.update_progress(progress_bar)

                _wait()

                # check completion
                response = refresh_query(*promise.get_action_ids())
                promise.update_data(self, response, progress_bar)

            progress_bar.update(max(0, progress_bar.total - progress_bar.n))

        promise.finalize()

    def build_refresh_query(self, field_type: GraphQLType) -> RefreshActionQuery:
        """Returns a RefreshActionQuery object for
        refreshing an object populated with response data.
        """
        return RefreshActionQuery(self._gql_client, field_type)


def create_environment(gql_client: Client) -> QctrlGraphQLEnvironment:
    """Creates the custom environment object with
    all custom handlers.
    """
    return QctrlGraphQLEnvironment(
        gql_client=gql_client,
        type_registry_cls=TypeRegistry,
        custom_handlers=[
            ComplexArrayEntryHandler,
            ComplexArrayHandler,
            ComplexArrayInputHandler,
            ComplexDenseArrayHandler,
            ComplexNumberHandler,
            ComplexNumberInputHandler,
            ComplexSparseArrayHandler,
            DateTimeScalarHandler,
            GraphScalarHandler,
            JsonDictScalarHandler,
            UnixTimeScalarHandler,
        ],
    )



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import numpy as np
from graphql import GraphQLList

from .base import BaseHandler
from .mixins import (
    DataMixin,
    PayloadMixin,
    SelectionMixin,
)


class ListHandler(BaseHandler, PayloadMixin, SelectionMixin, DataMixin):
    """Handler for GraphQLList types."""

    _field_type = GraphQLList

    def _format_payload(self, field_type, data):
        if not isinstance(data, (list, tuple, np.ndarray)):
            raise TypeError(
                f"{data} is of type {type(data)}, but expected a list, tuple, or array instead."
            )

        result = []

        for nested_data in data:
            result.append(self._env.format_payload(field_type.of_type, nested_data))

        return f"[ {', '.join(result)} ]"

    def _get_selection(self, field_type, overrides=None):
        return self._env.get_selection(field_type.of_type)

    def _load_data(self, field_type, data):
        result = []

        for nested_data in data:
            result.append(self._env.load_data(field_type.of_type, nested_data))

        return result

    def _update_data(self, field_type, obj, data):
        return self.load_data(field_type, data)
